# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
import numpy as np
#
class bspline(object):
    """Bspline class.

    Functions in the bspline library are implemented as methods on this
    class.

    Parameters
    ----------
    x
    nord
    npoly
    bkpt
    bkspread
    verbose

    Attributes
    ----------
    breakpoints
    nord
    npoly
    mask
    coeff
    icoeff
    xmin
    xmax
    funcname
    """
    def __init__(self,x,nord=4,npoly=1,bkpt=None,bkspread=1.0,verbose=False,**kwargs):
        """Init creates an object whose attributes are similar to the
        structure returned by the create_bspline function.
        """
        #
        # Set the breakpoints.
        #
        if bkpt is None:
            startx = x.min()
            rangex = x.max() - startx
            if 'placed' in kwargs:
                w = ((kwargs['placed'] >= startx) &
                    (kwargs['placed'] <= startx+rangex))
                if w.sum() < 2:
                    bkpt = np.arange(2,dtype='f')*rangex + startx
                else:
                    bkpt = kwargs['placed'][w]
            elif 'bkspace' in kwargs:
                nbkpts = int(rangex/kwargs['bkspace']) + 1
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts,dtype='f')*tempbkspace + startx
            elif 'nbkpts' in kwargs:
                nbkpts = kwargs['nbkpts']
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts,dtype='f')*tempbkspace + startx
            elif 'everyn' in kwargs:
                npkpts = max(nx/kwargs['everyn'], 1)
                if nbkpts == 1:
                    xspot = [0]
                else:
                    xspot = int(nx/(nbkpts-1))*np.arange(nbkpts,dtype='i4')
                bkpt = x[xspot].astype('f')
            else:
                raise ValueError('No information for bkpts.')
        imin = bkpt.argmin()
        imax = bkpt.argmax()
        if x.min() < bkpt[imin]:
            if verbose:
                print('Lowest breakpoint does not cover lowest x value: changing.')
            bkpt[imin] = x.min()
        if x.max() > bkpt[imax]:
            if verbose:
                print('Highest breakpoint does not cover highest x value: changing.')
            bkpt[imax] = x.max()
        nshortbkpt = bkpt.size
        fullbkpt = bkpt.copy()
        if nshortbkpt == 1:
            bkspace = np.float32(bkspread)
        else:
            bkspace = (bkpt[1] - bkpt[0]) * np.float32(bkspread)
        for i in np.arange(1,nord,dtype=np.float32):
            fullbkpt = np.insert(fullbkpt,0,bkpt[0]-bkspace*i)
            fullbkpt = np.insert(fullbkpt,fullbkpt.shape[0],
                bkpt[nshortbkpt-1] + bkspace*i)
        #
        # Set the attributes
        #
        nc = fullbkpt.size - nord
        self.breakpoints = fullbkpt
        self.nord = nord
        self.npoly = npoly
        self.mask = np.ones((fullbkpt.size,),dtype='bool')
        if npoly > 1:
            self.coeff = np.zeros((npoly,nc),dtype='d')
            self.icoeff = np.zeros((npoly,nc),dtype='d')
        else:
            self.coeff = np.zeros((nc,),dtype='d')
            self.icoeff = np.zeros((nc,),dtype='d')
        self.xmin = 0.0
        self.xmax = 1.0
        self.funcname = 'legendre'
        return
    #
    #
    #
    def fit(self,xdata,ydata,invvar,x2=None):
        """
        Parameters
        ----------
        xdata
        ydata
        invvar
        x2

        Returns
        -------
        fit
        """
        from . import cholesky_band, cholesky_solve
        goodbk = self.mask[self.nord:]
        nn = goodbk.sum()
        if nn < self.nord:
            yfit = np.zeros(ydata.shape,dtype='f')
            return (-2,yfit)
        nfull = nn * self.npoly
        bw = self.npoly * self.nord
        a1,lower,upper = self.action(xdata,x2=x2)
        foo = np.tile(invvar,bw).reshape(bw,invvar.size).transpose()
        a2 = a1 * foo
        alpha = np.zeros((bw,nfull+bw),dtype='d')
        beta = np.zeros((nfull+bw,),dtype='d')
        bi = np.arange(bw,dtype='i4')
        bo = np.arange(bw,dtype='i4')
        for k in range(1,bw):
            bi = np.append(bi, np.arange(bw-k,dtype='i4')+(bw+1)*k)
            bo = np.append(bo, np.arange(bw-k,dtype='i4')+bw*k)
        for k in range(nn-self.nord+1):
            itop = k*self.npoly
            ibottom = min(itop,nfull) + bw - 1
            ict = upper[k] - lower[k] + 1
            if ict > 0:
                work = np.dot(a1[lower[k]:upper[k]+1,:].T,a2[lower[k]:upper[k]+1,:])
                wb = np.dot(ydata[lower[k]:upper[k]+1],a2[lower[k]:upper[k]+1,:])
                alpha.T.flat[bo+itop*bw] += work.flat[bi]
                beta[itop:ibottom+1] += wb
        min_influence = 1.0e-10 * invvar.sum() / nfull
        errb = cholesky_band(alpha,mininf=min_influence) # ,verbose=True)
        if isinstance(errb[0],int) and errb[0] == -1:
            a = errb[1]
        else:
            yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
            return (self.maskpoints(errb),yfit)
        errs  = cholesky_solve(a,beta)
        if isinstance(errs[0],int) and errs[0] == -1:
            sol = errs[1]
        else:
            yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
            return (self.maskpoints(errs),yfit)
        if self.npoly > 1:
            self.icoeff[:,goodbk] = np.array(a[0,0:nfull].reshape(self.npoly,nn),dtype=a.dtype)
            self.coeff[:,goodbk] = np.array(sol[0:nfull].reshape(self.npoly,nn),dtype=sol.dtype)
        else:
            self.icoeff[goodbk] = np.array(a[0,0:nfull],dtype=a.dtype)
            self.coeff[goodbk] = np.array(sol[0:nfull],dtype=sol.dtype)
        yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
        return (0,yfit)
    #
    #
    #
    def action(self,x,x2=None):
        """
        Parameters
        ----------
        x
        x2

        Returns
        -------
        action
        """
        from ... import uniq
        from ...goddard.math import fchebyshev, flegendre
        nx = x.size
        nbkpt = self.mask.sum()
        if nbkpt < 2*self.nord:
            return (-2,0,0)
        n = nbkpt - self.nord
        gb = self.breakpoints[self.mask]
        bw = self.npoly*self.nord
        lower = np.zeros((n - self.nord + 1,),dtype='i4')
        upper = np.zeros((n - self.nord + 1,),dtype='i4') -1
        indx = self.intrv(x)
        bf1 = self.bsplvn(x, indx)
        action = bf1
        aa = uniq(indx,np.arange(indx.size,dtype='i4'))
        upper[indx[aa]-self.nord+1] = aa
        rindx = indx[::-1]
        bb = uniq(rindx,np.arange(rindx.size,dtype='i4'))
        lower[rindx[bb]-self.nord+1] = nx - bb - 1
        if x2 is not None:
            if x2.size != nx:
                raise ValueError('Dimensions of x and x2 do not match.')
            x2norm = 2.0 * (x2 - self.xmin) / (self.xmax - self.xmin ) - 1.0
            if self.funcname == 'poly':
                temppoly = np.ones((nx,self.npoly),dtype='f')
                for i in range(1,self.npoly):
                    temppoly[:,i] = temppoly[:,i-1] * x2norm
            elif self.funcname == 'poly1':
                temppoly = np.tile(x2norm,self.npoly).reshape(nx,self.npoly)
                for i in range(1,self.npoly):
                    temppoly[:,i] = temppoly[:,i-1] * x2norm
            elif self.funcname == 'chebyshev':
                temppoly = fchebyshev(x2norm,self.npoly)
            elif self.funcname == 'legendre':
                temppoly = flegendre(x2norm,self.npoly)
            else:
                raise ValueError('Unknown value of funcname.')
            action = np.zeros((nx,bw),dtype='d')
            counter = -1
            for ii in range(self.nord):
                for jj in range(self.npoly):
                    counter += 1
                    action[:,counter] = bf1[:,ii]*temppoly[:,jj]
        return (action,lower,upper)
    #
    #
    #
    def intrv(self,x):
        """
        Parameters
        ----------
        x

        Returns
        -------
        intrv
        """
        gb = self.breakpoints[self.mask]
        n = gb.size - self.nord
        indx = np.zeros((x.size,),dtype='i4')
        ileft = self.nord -1
        for i in range(x.size):
            while x[i] > gb[ileft+1] and ileft < n - 1:
                ileft += 1
            indx[i] = ileft
        return indx
    #
    #
    #
    def bsplvn(self,x,ileft):
        """
        Parameters
        ----------
        x
        ileft

        Returns
        -------
        bsplvn
        """
        bkpt = self.breakpoints[self.mask]
        vnikx = np.zeros((x.size,self.nord),dtype=x.dtype)
        deltap = vnikx.copy()
        deltam = vnikx.copy()
        j = 0
        vnikx[:,0] = 1.0
        while j < self.nord - 1:
            ipj = ileft+j+1
            deltap[:,j] = bkpt[ipj] - x
            imj = ileft-j
            deltam[:,j] = x - bkpt[imj]
            vmprev = 0.0
            for l in range(j+1):
                vm = vnikx[:,l]/(deltap[:,l] + deltam[:,j-l])
                vnikx[:,l] = vm*deltap[:,l] + vmprev
                vmprev = vm*deltam[:,j-l]
            j += 1
            vnikx[:,j] = vmprev
        return vnikx
    #
    #
    #
    def value(self,x,x2=None,action=None,lower=None,upper=None):
        """
        Parameters
        ----------
        x
        x2
        action
        lower
        upper

        Returns
        -------
        value
        """
        xsort = x.argsort()
        xwork = x[xsort]
        if x2 is not None:
            x2work = x2[xsort]
        else:
            x2work = None
        if action is not None:
            if lower is None or upper is None:
                raise ValueError('Must specify lower and upper if action is set.')
        else:
            action,lower,upper = self.action(xwork,x2=x2work)
        yfit = np.zeros(x.shape,dtype=x.dtype)
        bw = self.npoly * self.nord
        spot = np.arange(bw,dtype='i4')
        goodbk = self.mask.nonzero()[0]
        coeffbk = self.mask[self.nord:].nonzero()[0]
        n = self.mask.sum() - self.nord
        if self.npoly > 1:
            goodcoeff = self.coeff[:,coeffbk]
        else:
            goodcoeff = self.coeff[coeffbk]
        # maskthis = np.zeros(xwork.shape,dtype=xwork.dtype)
        for i in range(n-self.nord+1):
            ict = upper[i] - lower[i] + 1
            if ict > 0:
                yfit[lower[i]:upper[i]+1] = np.dot(
                    action[lower[i]:upper[i]+1,:],goodcoeff[i*self.npoly+spot])
        yy = yfit.copy()
        yy[xsort] = yfit
        mask = np.ones(x.shape,dtype='bool')
        gb = self.breakpoints[goodbk]
        outside = ((x < gb[self.nord-1]) |
            (x > gb[n]))
        if outside.any():
            mask[outside] = False
        hmm = ((np.diff(goodbk) > 2).nonzero())[0]
        for jj in range(hmm.size):
            inside = ((x >= self.breakpoints[goodbk[hmm[jj]]]) &
                (x <= self.breakpoints[goodbk[hmm[jj]+1]-1]))
            if inside.any():
                mask[inside] = False
        return (yy,mask)
    #
    #
    #
    def maskpoints(self,err):
        """
        Parameters
        ----------
        err

        Returns
        -------
        maskpoints
        """
        nbkpt = self.mask.sum()
        if nbkpt <= 2*self.nord:
            return -2
        hmm = err[np.unique(errb/self.npoly)]/self.npoly
        n = nbkpt - self.nord
        if np.any(hmm >= n):
            return -2
        test = np.zeros(nbkpt,dtype='bool')
        for jj in range(-np.ceil(nord/2.0),nord/2.0):
            foo = np.where((hmm+jj) > 0,hmm+jj,np.zeros(hmm.shape,dtype=hmm.dtype))
            inside = np.where((foo+nord) < n-1,foo+nord,np.zeros(hmm.shape,dtype=hmm.dtype)+n-1)
            test[inside] = True
        if test.any():
            reality = self.mask[test]
            if self.mask[reality].any():
                self.mask[reality] = False
                return -1
            else:
                return -2
        else:
            return -2




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the misc directory of idlutils.
"""
from .decode_mixed import decode_mixed
from .djs_laxisnum import djs_laxisnum
#from .djs_readcol import djs_readcol
from .hogg_iau_name import hogg_iau_name
from .struct_print import struct_print



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_filtername():
    from .. import filtername
    assert filtername(0) == 'u'
    assert filtername(1) == 'g'
    assert filtername(2) == 'r'
    assert filtername(3) == 'i'
    assert filtername(4) == 'z'
    #
    # filtername should return its argument if it's not
    # integer-like
    #
    assert filtername('r') == 'r'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def default_skyversion():
    """Returns skyversion number to use for photoop imaging.

    Returns
    -------
    default_skyversion : int
        The default skyversion number.

    Notes
    -----
    The skyversion number is hard-coded to 2.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import default_skyversion
    >>> default_skyversion()
    2
    """
    return 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def spherematch(ra1,dec1,ra2,dec2,matchlength,chunksize=None,maxmatch=1,debug=False):
    """Match points on a sphere.

    Parameters
    ----------
    ra1, dec1, ra2, dec2 : ndarray
        The sets of coordinates to match.  Assumed to be in decimal degrees
    matchlength : float
        Two points closer than this separation are matched. Assumed to be in decimal degrees.
    chunksize : float, optional
        Value to pass to chunk assignment.
    maxmatch : int, optional
        Allow up to `maxmatch` matches per coordinate.  Default 1. If set to zero,
        All possible matches will be returned.
    debug : bool, optional

    Returns
    -------
    spherematch : tuple
        A tuple containing the indices into the first set of points, the
        indices into the second set of points and the match distance in
        decimal degrees.
    """
    import numpy as np
    from . import chunks
    from ...goddard.astro import gcirc
    #
    # Set default values
    #
    if chunksize is None:
        chunksize = max(4.0*matchlength,0.1)
    #
    # Initialize chunks
    #
    chunk = chunks(ra1,dec1,chunksize)
    chunk.assign(ra2,dec2,matchlength)
    if debug:
        print("raOffset = {0:7.3f}.".format(chunk.raOffset))
        print(chunk.raBounds)
        print(chunk.decBounds)
        print(chunk.chunkList)
    #
    # Create return arrays
    #
    match1 = list()
    match2 = list()
    distance12 = list()
    for i in range(ra1.size):
        currra = np.fmod(ra1[i]+chunk.raOffset,360.0)
        rachunk,decchunk = chunk.get(currra,dec1[i])
        jmax = len(chunk.chunkList[decchunk][rachunk])
        if jmax > 0:
            for j in range(jmax):
                k = chunk.chunkList[decchunk][rachunk][j]
                sep = gcirc(ra1[i],dec1[i],ra2[k],dec2[k],units=2)/3600.0
                if sep < matchlength:
                    match1.append(i)
                    match2.append(k)
                    distance12.append(sep)
    #
    # Sort distances
    #
    omatch1 = np.array(match1)
    omatch2 = np.array(match2)
    odistance12 = np.array(distance12)
    s = odistance12.argsort()
    #
    # Retain only desired matches
    #
    if maxmatch > 0:
        gotten1 = np.zeros(ra1.size,dtype='i4')
        gotten2 = np.zeros(ra2.size,dtype='i4')
        nmatch = 0
        for i in range(omatch1.size):
            if (gotten1[omatch1[s[i]]] < maxmatch and
                gotten2[omatch2[s[i]]] < maxmatch):
                gotten1[omatch1[s[i]]] += 1
                gotten2[omatch2[s[i]]] += 1
                nmatch += 1
        match1 = np.zeros(nmatch,dtype='i4')
        match2 = np.zeros(nmatch,dtype='i4')
        distance12 = np.zeros(nmatch,dtype='d')
        gotten1[:] = 0
        gotten2[:] = 0
        nmatch = 0
        for i in range(omatch1.size):
            if (gotten1[omatch1[s[i]]] < maxmatch and
                gotten2[omatch2[s[i]]] < maxmatch):
                gotten1[omatch1[s[i]]] += 1
                gotten2[omatch2[s[i]]] += 1
                match1[nmatch] = omatch1[s[i]]
                match2[nmatch] = omatch2[s[i]]
                distance12[nmatch] = odistance12[s[i]]
                nmatch += 1
    else:
        match1 = omatch1[s]
        match2 = omatch2[s]
        distance12 = odistance12[s]
    return (match1,match2,distance12)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_astrombad(run,camcol,field,photolog_version='dr10'):
    """For a list of RUN, CAMCOL, FIELD, return whether each field has bad astrometry.

    Parameters
    ----------
    run, camcol, field : int or array of int
        Run, camcol and field.  If arrays are passed,
        all must have the same length.
    photolog_version : str, optional
        Use this version of photolog to obtain the sdssMaskbits.par file,
        if $PHOTOLOG_DIR is not set.

    Returns
    -------
    sdss_astrombad : ndarray of bool
        Array of bool.  ``True`` indicates the field is bad.

    Raises
    ------
    ValueError
        If the sizes of the arrays don't match or if the array values are out of bounds.

    Notes
    -----
    Reads data from ``$PHOTOLOG_DIR/opfiles/opBadFields.par``.

    If there is a problem with one camcol, we assume a
    problem with all camcols.

    """
    from . import opbadfields
    from ..yanny import yanny
    from numpy import array, bool, int64, zeros
    from astropy.utils.data import download_file
    from os import getenv
    from os.path import join
    #
    # Check inputs
    #
    if isinstance(run,int):
        #
        # Assume all inputs are integers & promote to arrays.
        #
        run = array([run],dtype=int64)
        camcol = array([camcol],dtype=int64)
        field = array([field],dtype=int64)
    else:
        #
        # Check that all inputs have the same shape.
        #
        if run.shape != camcol.shape:
            raise ValueError("camcol.shape does not match run.shape!")
        if run.shape != field.shape:
            raise ValueError("field.shape does not match run.shape!")
    #
    # Check ranges of parameters
    #
    if ((run < 0) | (run >= 2**16)).any():
        raise ValueError("run values are out-of-bounds!")
    if ((camcol < 1) | (camcol > 6)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((field < 0) | (field >= 2**12)).any():
        raise ValueError("camcol values are out-of-bounds!")
    #
    # Read the file
    #
    if opbadfields is None:
        if getenv('PHOTOLOG_DIR') is None:
            if photolog_version == 'trunk' or photolog_version.startswith('branches/'):
                iversion = photolog_version
            else:
                iversion = 'tags/'+photolog_version
            baseurl = 'http://www.sdss3.org/svn/repo/photolog/{0}/opfiles/opBadfields.par'.format(iversion)
            filename = download_file(baseurl,cache=True)
        else:
            filename = join(getenv('PHOTOLOG_DIR'),'opfiles','opBadfields.par')
        astrombadfile = yanny(filename,np=True)
        w = ((astrombadfile['BADFIELDS']['problem'] == 'astrom'.encode()) |
            (astrombadfile['BADFIELDS']['problem'] == 'rotator'.encode()))
        opbadfields = astrombadfile['BADFIELDS'][w]
    #
    # opbadfields already has astrom problems selected at this point
    #
    bad = zeros(run.shape,dtype=bool)
    for row in opbadfields:
        w = ((run == row['run']) &
        (field >= row['firstfield']) & (field < row['lastfield']))
        if w.any():
            bad[w] = True
    return bad



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def aesthetics(flux,invvar,method='traditional'):
    """Add nice values to a spectrum where it is masked.

    Parameters
    ----------
    flux : ndarray
        The spectrum to clean up.
    invvar : ndarray
        Inverse variance of the spectrum.
    method : { 'traditional', 'noconst', 'mean', 'damp', 'nothing' }, optional
        Apply this method to clean up the spectrum.  Default is 'traditional'.

    Returns
    -------
    aesthetics : ndarray
        A cleaned-up spectrum.
    """
    import numpy as np
    from scipy.special import erf
    from ...pydlutils.image import djs_maskinterp
    from .. import Pydlspec2dException
    badpts = invvar == 0
    if badpts.any():
        if method == 'traditional':
            newflux = djs_maskinterp(flux,invvar == 0,const=True)
        elif method == 'noconst':
            newflux = djs_maskinterp(flux,invvar == 0)
        elif method == 'mean':
            newflux = flux.copy()
            goodpts = invvar > 0
            newflux[~goodpts] = newflux[goodpts].mean()
        elif method == 'damp':
            l = 250 # damping length in pixels
            goodpts = invvar.nonzero()[0]
            nflux = flux.size
            mingood = goodpts.min()
            maxgood = goodpts.max()
            newflux = djs_maskinterp(flux,invvar == 0,const=True)
            pixels = np.arange(nflux,dtype='f')
            if mingood > 0:
                damp1 = float(min(mingood,l))
                newflux *= 0.5*(1.0+erf((pixels-mingood)/damp1))
            if maxgood < (nflux - 1):
                damp2 = float(min(maxgood,l))
                newflux *= 0.5*(1.0+erf((maxgood-pixels)/damp2))
        elif method == 'nothing':
            newflux = flux.copy()
        else:
            raise Pydlspec2dException("Unknown method: %s" % method)
        return newflux
    else:
        return flux




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def readcol(name,**kwargs):
    """This is a placeholder for the real Goddard readcol
    function.  Eventually this will match the readcol functionality.
    """
    from ...pydlutils.misc import djs_readcol
    return djs_readcol(name,**kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def fchebyshev(x,m):
    """Compute a Chebyshev polynomial.

    Parameters
    ----------
    x : array_like
    m : int

    Returns
    -------
    fchebyshev : array_like
    """
    import numpy as np
    from scipy.special import chebyt
    if isinstance(x,np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 1:
        raise ValueError('Order of Chebyshev polynomial must be at least 1.')
    leg = np.ones((m,n),dtype='d')
    if m >= 2:
        leg[1,:] = x
    if m >= 3:
        for k in range(2,m):
            leg[k,:] = np.polyval(chebyt(k),x)
    return leg




# Licensed under a 3-clause BSD style license
# -*- coding: utf-8 -*-
def get_package_data():
    # Installs the testing data files.  Unable to get package_data
    # to deal with a directory hierarchy of files, so just explicitly list.
    return {
        'pydl.pydlutils.yanny.tests': ['t/*',]
    }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def write_ndarray_to_yanny(filename,datatables,structnames=None,
                           enums=None,hdr=None,comments=None):
    """Converts a NumPy record array into a new FTCL/yanny file.

    Returns a new yanny object corresponding to the file.

    Parameters
    ----------
    filename : str
        The name of a parameter file.
    datatables : numpy.ndarray or list of numpy.ndarray
        A NumPy record array containing data that can be copied into a yanny object.
    structnames : str or list of str, optional
        The name(s) to give the structure(s) in the yanny file.  Defaults to 'MYSTRUCT0'.
    enums : dict, optional
        A dictionary containing enum information.  See the documentation for
        the `dtype_to_struct` method of the yanny object.
    hdr : dict, optional
        A dictionary containing keyword/value pairs for the 'header' of the yanny file.
    comments : str or list of str, optional
        A string containing comments that will be added to the start of the new file.

    Returns
    -------
    par : pydl.pydlutils.yanny.yanny
        The `yanny` object resulting from writing the file.

    Examples
    --------
    """
    from numpy import ndarray
    from . import yanny
    from .. import PydlutilsException
    par = yanny(filename,np=True)
    if par:
        #
        # If the file already exists
        #
        raise PydlutilsException("Apparently {0} already exists.".format(filename))
    if type(datatables) == ndarray:
        datatables = (datatables,)
    if structnames is None:
        structnames = ["MYSTRUCT{0:d}".format(k) for k in range(len(datatables))]
    if isinstance(structnames,str) or isinstance(structnames,unicode):
        structnames = (structnames,)
    if len(datatables) != len(structnames):
        raise PydlutilsException("The data tables and their names do not match!")
    for k in range(len(datatables)):
        struct = par.dtype_to_struct(datatables[k].dtype,structname=structnames[k],enums=enums)
        par['symbols']['struct'] += struct['struct']
        par['symbols'][structnames[k].upper()] = struct[structnames[k].upper()]
        if enums is not None and len(par['symbols']['enum']) == 0:
            par['symbols']['enum'] = struct['enum']
        par[structnames[k].upper()] = datatables[k]
    if hdr is not None:
        for key in hdr:
            par[key] = hdr[key]
    par.write(filename,comments=comments)
    return par



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
====
pydl
====

Python replacements for functions that are part of the IDL_ built-in library, or
part of astronomical IDL libraries.  The emphasis is on reproducing results of
the astronomical library functions.  Only the bare minimum of IDL_ built-in
functions are implemented to support this.

.. _IDL: http://www.exelisvis.com/language/en-us/productsservices/idl.aspx
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

#
# pydl stuff starts here
#
from .file_lines import file_lines
from .pcomp import pcomp
from .smooth import smooth
from .uniq import uniq
from . import goddard
from . import photoop
from . import pydlutils
from . import pydlspec2d

class PydlException(Exception):
    pass

__all__ = ['file_lines', 'pcomp', 'smooth', 'uniq', 'PydlException']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage implements functions from the idlutils package.
"""
#
# Define this early on so that submodules can use it
#
#from .. import PydlException
#class PydlutilsException(PydlException):
#    pass
class PydlutilsException(Exception):
    pass
#
# Import subpackages
#
from . import bspline
from . import cooling
#from . import coord
#from . import fits
from . import image
from . import mangle
from . import math
from . import misc
from . import sdss
from . import spheregroup
from . import yanny
#
# Set __all__
#
__all__ = ['PydlutilsException', 'bspline', 'cooling', 'image', 'mangle',
    'math', 'misc', 'sdss', 'spheregroup', 'yanny']



#
#
#
def angles_to_xyz(r,phi,theta):
    """Convert spherical coordinates (r,phi,theta) into Cartesion coordinates

    This uses the standard mathematical definition of spherical coordinates.
    To convert e.g. RA, Dec or longitude, latitude, pass RA, 90 - Dec.
    """
    import numpy as np
    x = r * np.cos(np.deg2rad(phi)) * np.sin(np.deg2rad(theta))
    y = r * np.sin(np.deg2rad(phi)) * np.sin(np.deg2rad(theta))
    z = r * np.cos(np.deg2rad(theta))
    return (x,y,z)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def djs_laxisnum(dims,iaxis=0):
    """Returns an integer array where each element of the array is set equal to its index number in the specified axis.

    Parameters
    ----------
    dims : list
        Dimensions of the array to return.
    iaxis : int, optional
        Index along this dimension.

    Returns
    -------
    djs_laxisnum : ndarray
        An array of indexes with ``dtype=int32``.

    Raises
    ------
    ValueError
        If `iaxis` is greater than or equal to the number of dimensions.
    NotImplementedError
        If the number of dimensions is greater than three.

    Examples
    --------
    >>> from pydl.pydlutils.misc import djs_laxisnum
    >>> djs_laxisnum([4,4])
    array([[0, 0, 0, 0],
           [1, 1, 1, 1],
           [2, 2, 2, 2],
           [3, 3, 3, 3]], dtype=int32)
    """
    from numpy import zeros
    ndimen = len(dims)
    result = zeros(dims,dtype='i4')
    if ndimen == 1:
        pass
    elif ndimen == 2:
        if iaxis == 0:
            for k in range(dims[0]):
                result[k,:] = k
        elif iaxis == 1:
            for k in range(dims[1]):
                result[:,k] = k
        else:
            raise ValueError("Bad value for iaxis: {0:d}".format(iaxis))
    elif ndimen == 3:
        if iaxis == 0:
            for k in range(dims[0]):
                result[k,:,:] = k
        elif iaxis == 1:
            for k in range(dims[1]):
                result[:,k,:] = k
        elif iaxis == 2:
            for k in range(dims[2]):
                result[:,:,k] = k
        else:
            raise ValueError("Bad value for iaxis: {0:d}".format(iaxis))
    else:
        raise NotImplementedError("{0:d} dimensions not supported.".format(ndimen))
    return result




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def iterfit(xdata,ydata,invvar=None,upper=5,lower=5,x2=None,maxiter=10,**kwargs):
    """Iteratively fit a bspline set to data.

    Parameters
    ----------
    xdata
    ydata
    invvar
    upper
    lower
    x2
    maxiter

    Returns
    -------
    iterfit : tuple
    """
    import numpy as np
    from .bspline import bspline
    from ..math import djs_reject
    nx = xdata.size
    if ydata.size != nx:
        raise ValueError('Dimensions of xdata and ydata do not agree.')
    if invvar is not None:
        if invvar.size != nx:
            raise ValueError('Dimensions of xdata and invvar do not agree.')
    else:
        #
        # This correction to the variance makes it the same
        # as IDL's variance()
        #
        var = ydata.var()*(float(nx)/float(nx-1))
        if var == 0:
            var = 1.0
        invvar = np.ones(ydata.shape,dtype=ydata.dtype)/var
    if x2 is not None:
        if x2.size != nx:
            raise ValueError('Dimensions of xdata and x2 do not agree.')
    yfit = np.zeros(ydata.shape)
    if invvar.size == 1:
        outmask = True
    else:
        outmask = np.ones(invvar.shape,dtype='bool')
    xsort = xdata.argsort()
    maskwork = (outmask & (invvar > 0))[xsort]
    if 'oldset' in kwargs:
        sset = kwargs['oldset']
        sset.mask = True
        sset.coeff = 0
    else:
        if not maskwork.any():
            raise ValueError('No valid data points.')
            #return (None,None)
        if 'fullbkpt' in kwargs:
            fullbkpt = kwargs['fullbkpt']
        else:
            sset = bspline(xdata[xsort[maskwork]],**kwargs)
            if maskwork.sum() < sset.nord:
                print('Number of good data points fewer than nord.')
                return (sset,outmask)
            if x2 is not None:
                if 'xmin' in kwargs:
                    xmin = kwargs['xmin']
                else:
                    xmin = x2.min()
                if 'xmax' in kwargs:
                    xmax = kwargs['xmax']
                else:
                    xmax = x2.max()
                if xmin == xmax:
                    xmax = xmin + 1
                sset.xmin = xmin
                sset.xmax = xmax
                if 'funcname' in kwargs:
                    sset.funcname = kwargs['funcname']
    xwork = xdata[xsort]
    ywork = ydata[xsort]
    invwork = invvar[xsort]
    if x2 is not None:
        x2work = x2[xsort]
    else:
        x2work = None
    iiter = 0
    error = 0
    qdone = -1
    while (error != 0 or qdone == -1) and iiter <= maxiter:
        goodbk = sset.mask.nonzero()[0]
        if maskwork.sum() <= 1 or not sset.mask.any():
            sset.coeff = 0
            iiter = maxiter + 1
        else:
            if 'requiren' in kwargs:
                i = 0
                while xwork[i] < sset.breakpoints[goodbk[sset.nord]] and i < nx-1:
                    i +=1
                ct = 0
                for ileft in range(sset.nord,sset.mask.sum()-sset.nord+1):
                    while (xwork[i] >= sset.breakpoints[goodbk[ileft]] and
                        xwork[i] < sset.breakpoints[goodbk[ileft+1]] and
                        i < nx-1):
                        ct += invwork[i]*maskwork[i] > 0
                        i += 1
                    if ct >= kwargs['requiren']:
                        ct = 0
                    else:
                        sset.mask[goodbk[ileft]] = False
            error,yfit = sset.fit(xwork,ywork, invwork*maskwork,
                x2=x2work)
        iiter += 1
        inmask = maskwork
        if error == -2:
            return (sset,outmask)
        elif error == 0:
            maskwork,qdone = djs_reject(ywork, yfit, invvar=invwork, inmask=inmask,
                outmask=maskwork, upper=upper, lower=lower)
        else:
            pass
    outmask[xsort] = maskwork
    temp = yfit
    yfit[xsort] = temp
    return (sset,outmask)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_path(ftype, run, camcol=0, rerun=''):
    """Return the path name for SDSS data assuming SAS directory structure.

    Parameters
    ----------
    ftype : str
        The general type of the file, for example ``'reObj'``
    run : int
        The run number.
    camcol : int, optional
        If necessary, set the camcol number using this argument.
    rerun : str, optional
        If necessary, set the rerun name using this argument.

    Returns
    -------
    sdss_name : str
        The directory in which file `ftype` lives.

    Raises
    ------
    KeyError
        If the file type is unknown.

    """
    from os import getenv
    from . import _path_formats
    indict = {
        'run':run,
        'camcol':camcol,
        'rerun':rerun,
        'calib':getenv('PHOTO_CALIB'),
        'data':getenv('PHOTO_DATA'),
        'photoobj':getenv('BOSS_PHOTOOBJ'),
        'redux':getenv('PHOTO_REDUX'),
        'resolve':getenv('PHOTO_RESOLVE'),
        'sky':getenv('PHOTO_SKY'),
        'sweep':getenv('PHOTO_SWEEP'),
        }
    try:
        datadir = _path_formats[ftype].format(**indict)
    except KeyError:
        raise KeyError("Unknown FTYPE = {0}".format(ftype))
    return datadir



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_flagname(flagname, flagvalue, concat=False):
    """Return a list of flag names corresponding to the values.

    Parameters
    ----------
    flagname : str
        The name of a bitmask group. Not case-sensitive.
    flagvalue : long
        The value to be converted into bitmask names.
    concat : bool, optional
        If set to ``True``, the list of names is converted to a space-separated string.

    Returns
    -------
    sdss_flagname : str or list
        The names of the bitmasks encoded in `flagvalue`.

    Raises
    ------
    KeyError
        If `flagname` is invalid

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_flagname
    >>> sdss_flagname('ANCILLARY_TARGET1',2310346608843161600) # doctest: +REMOTE_DATA
    ['BRIGHTGAL', 'BLAZGX', 'ELG']
    """
    from . import maskbits
    if maskbits is None:
        from .set_maskbits import set_maskbits
        maskbits = set_maskbits()
    flagu = flagname.upper()
    bits = list()
    for bit in range(64):
        if (flagvalue & (1 << bit)) != 0:
            bits.append(bit)
    retval = list()
    for bit in bits:
        try:
            f = [ x for x in maskbits[flagu].items() if x[1] == bit ]
        except KeyError:
            raise KeyError("Unknown flag group {0}!".format(flagu))
        if f:
            retval.append(f[0][0])
    if concat:
        retval = ' '.join(retval)
    return retval




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_findspec():
    """This is just a placeholder for now."""
    from .. import findspec
    #slist = findspec(infile='file.in',sdss=True)
    #if slist is not None:
    #    print slist
    assert True
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage implements functions from the photoop package.
"""
#
# Define this early on so that submodules can use it
#
#from .. import PydlException
#class PhotoopException(PydlException):
#    pass
class PhotoopException(Exception):
    pass
#
# Import subpackages
#
from . import photoobj
from . import sdssio
#
# Set __all__
#
__all__ = ['PhotoopException', 'photoobj', 'sdssio' ]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_flagval(flagname,bitname):
    """Convert bitmask names into values.

    Converts human-readable bitmask names into numerical values.  The inputs
    are not case-sensitive; all inputs are converted to upper case internally.

    Parameters
    ----------
    flagname : str
        The name of a bitmask group.
    bitname : str or list
        The name(s) of the specific bitmask(s) within the `flagname` group.

    Returns
    -------
    sdss_flagval : numpy.uint64
        The value of the bitmask name(s).

    Raises
    ------
    KeyError
        If `flagname` or `bitname` are invalid names.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_flagval
    >>> sdss_flagval('ANCILLARY_TARGET1',['BLAZGX','ELG','BRIGHTGAL']) # doctest: +REMOTE_DATA
    2310346608843161600
    """
    from numpy import uint64
    from . import maskbits
    if maskbits is None:
        from .set_maskbits import set_maskbits
        maskbits = set_maskbits()
    #
    # Make sure inlabel is a list
    #
    if isinstance(bitname,str):
        bitnames = [bitname.upper()]
    else:
        bitnames = [b.upper() for b in bitname]
    flagu = flagname.upper()
    flagvalue = uint64(0)
    for bit in bitnames:
        if flagu in maskbits:
            if bit in maskbits[flagu]:
                flagvalue += uint64(2)**uint64(maskbits[flagu][bit])
            else:
                raise KeyError("Unknown bit label {0} for flag group {1}!".format(bit, flagu))
        else:
            raise KeyError("Unknown flag group {0}!".format(flagu))
    return flagvalue



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
def test_file_lines():
    from ..file_lines import file_lines
    from os.path import basename, dirname, join
    import glob
    #
    # Find the test files
    #
    datadir = join(dirname(__file__),'t')
    fileglob = join(datadir,'this-file-contains-*-lines.txt')
    plainfiles = glob.glob(fileglob)
    gzfiles = glob.glob(fileglob+'.gz')
    for p in plainfiles:
        n = file_lines(p)
        number_of_lines = int(basename(p).split('-')[3])
        assert n == number_of_lines
    for p in gzfiles:
        n = file_lines(p,compress=True)
        number_of_lines = int(basename(p).split('-')[3])
        assert n == number_of_lines
    #
    # Test list passing
    #
    n = file_lines(plainfiles)
    number_of_lines = [int(basename(p).split('-')[3]) for p in plainfiles]
    assert n == number_of_lines
    n = file_lines(gzfiles,compress=True)
    number_of_lines = [int(basename(p).split('-')[3]) for p in gzfiles]
    assert n == number_of_lines
    #
    # Make sure empty files work
    #
    n = file_lines(join(datadir,'this-file-is-empty.txt'))
    assert n == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_spherematch():
    import numpy as np
    from .. import spherematch
    i1_should_be = np.array([17,  0,  2, 16, 12, 13,  1,  5, 15,  7, 19,  8, 11, 10, 14, 18,  3,
                              9,  6,  4])
    i2_should_be = np.array([ 2,  0, 17,  3, 16, 15, 14,  5,  6, 10,  8, 19, 18,  4,  9,  7, 11,
                              12,  1, 13])
    np.random.seed(137)
    searchrad = 3.0/3600.0
    n = 20
    ra1 = 360.0*np.random.random((n,))
    dec1 = 90.0 - np.rad2deg(np.arccos(2.0*np.random.random((n,)) - 1.0))
    ra2 = ra1 + np.random.normal(0,1.0/3600.0)
    dec2 = dec1 + np.random.normal(0,1.0/3600.0)
    foo = np.arange(n)
    np.random.shuffle(foo)
    i1, i2, d12 = spherematch(ra1, dec1, ra2[foo], dec2[foo], searchrad, maxmatch=0)
    assert (i1 == i1_should_be).all()
    assert (i2 == i2_should_be).all()




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the yanny directory of idlutils.
"""
from .write_ndarray_to_yanny import write_ndarray_to_yanny
from .yanny import yanny



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def struct_print(array,filename=None,formatcodes=None,alias=None,fdigit=5,ddigit=7,
    html=False,no_head=False,debug=False,silent=False):
    """Print a NumPy record array (analogous to an IDL structure) in a nice way.

    Parameters
    ----------
    array : ndarray
        A record array to print.
    filename : str or file-like, optional
        If supplied, write to this file.
    formatcodes : dict, optional
        If supplied, use explicit format for certain columns.
    alias : dict, optional
        If supplied, use this mapping of record array column names to printed column names.
    fdigit : int, optional
        Width of 32-bit floating point columns, default 5.
    ddigit : int, optional
        Width of 64-bit floating point columns, default 7.
    html : bool, optional
        If ``True``, print an html table.
    no_head : bool, optional
        If ``True``, *don't* print a header line.
    debug : bool, optional
        If ``True``, print some extra debugging information.
    silent : bool, optional
        If ``True``, do not print the table, just return it.
    Returns
    -------
    struct_print : tuple
        A tuple containing a list of the lines in the table.  If `html` is ``True``,
        also returns a list of lines of CSS for formatting the table.

    Examples
    --------
    >>> import numpy as np
    >>> from pydl.pydlutils.misc import struct_print
    >>> struct_print(np.array([(1,2.34,'five'),(2,3.456,'seven'),(3,4.5678,'nine')],dtype=[('a','i4'),('bb','f4'),('ccc','S5')]),silent=True)
    (['a bb          ccc  ', '- ----------- -----', '1        2.34 five ', '2       3.456 seven', '3      4.5678 nine '], [])
    """
    import numpy as np
    from . import decode_mixed
    from .. import PydlutilsException
    f = None # This variable will store a file handle
    if filename is not None:
        if isinstance(filename,file):
            f = filename
        else:
            f = open(filename,'w')
    if html:
        headstart = '<tr><th>'
        headsep = '</th><th>'
        headend = '</th></tr>'
        colstart = '<tr><td>'
        colsep = '</td><td>'
        colend = '</td></tr>'
        css = [ '<style type="text/css">',
            'table {',
            '    border-collapse: collapse;',
            '}',
            'th {',
            '    padding: 2px;',
            '    text-align: right;',
            '    border: 1px solid black;',
            '    font-weight: bold;',
            '}',
            'td {',
            '    padding: 2px;',
            '    text-align: right;',
            '    border: 1px solid black;',
            '}',
            '</style>' ]
    else:
        headstart = ''
        headsep = ' '
        headend = ''
        colstart = ''
        colsep = ' '
        colend = ''
        css = list()
    #
    # Alias should be a dictionary that maps structure names to column names
    #
    if alias is None:
        #
        # Create a dummy alias dictionary
        #
        alias = dict(list(zip(array.dtype.names,array.dtype.names)))
    else:
        #
        # Fill in any missing values of the alias dictionary
        #
        for tag in array.dtype.names:
            if tag not in alias:
                alias[tag] = tag
    #
    # Formatcodes allows an override for certain columns.
    #
    if formatcodes is None:
        formatcodes = dict()
    #
    # This dictionary will hold the number of characters in each column
    #
    nchar = dict()
    #
    # Construct format codes for each column
    #
    for k, tag in enumerate(array.dtype.names):
        if tag in formatcodes:
            thiscode = formatcodes[tag]
            thisn = len(thiscode.format(array[tag][0]))
        else:
            d = array.dtype.fields[tag][0]
            if d.kind == 'i':
                thisn = max(max(len(str(array[tag].min())),len(str(array[tag].max()))),len(tag))
                thiscode = "{{{0:d}:{1:d}d}}".format(k,thisn)
            elif d.kind == 'f':
                if d.itemsize == 8:
                    prec = ddigit
                else:
                    prec = fdigit
                thisn = prec + 6
                if array[tag].min() < 0:
                    thisn += 1
                thiscode = "{{{0:d}:{1:d}.{2:d}g}}".format(k,thisn,prec)
            elif d.kind == 'S':
                thisn = max(d.itemsize,len(tag))
                thiscode = "{{{0:d}:{1:d}s}}".format(k,thisn)
            else:
                raise PydlutilsException("Unsupported kind: {0}".format(d.kind))
            formatcodes[tag] = thiscode
        nchar[tag] = thisn
    #
    # Start building an array of lines
    #
    lines = list()
    #
    # Construct header lines
    #
    if html:
        lines.append('<table>')
        hdr1 = headstart + headsep.join([alias[tag] for tag in array.dtype.names]) + headend
        lines.append(hdr1)
    else:
        if not no_head:
            hdr1 = headstart + headsep.join([("{{0:{0:d}s}}".format(nchar[tag])).format(alias[tag]) for tag in array.dtype.names]) + headend
            hdr2 = headstart + headsep.join(['-' * nchar[tag] for tag in array.dtype.names]) + headend
            lines.append(hdr1)
            lines.append(hdr2)
    #
    # Create a format string for the data from the individual format codes
    #
    rowformat = colstart + colsep.join([formatcodes[tag] for tag in array.dtype.names]) + colend
    if debug:
        print(rowformat)
    for k in range(array.size):
        lines.append(rowformat.format(*([decode_mixed(l) for l in array[k].tolist()])))
    if html:
        lines.append('</table>')
    if f is None:
        if not silent:
            print("\n".join(lines)+"\n")
    else:
        f.write("\n".join(lines)+"\n")
        f.close()
    return (lines,css)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def number_of_fibers(plate,**kwargs):
    """Returns the total number of fibers per plate.

    Parameters
    ----------
    plate : int or ndarray
        The plate(s) to examine.

    Returns
    -------
    number_of_fibers : ndarray
        The number of fibers on each plate.
    """
    import os
    import os.path
    from astropy.io import fits as pyfits
    import numpy as np
    from . import latest_mjd
    #
    # Get mjd values
    #
    mjd = latest_mjd(plate,**kwargs)
    nfiber = np.zeros(mjd.size,dtype='i4')
    #
    # SDSS-I,II plates
    #
    nfiber[mjd < 55025] = 640
    #
    # Short circuit if we're done.
    #
    if (nfiber == 640).all():
        return nfiber
    #
    # Not all BOSS plates have 1000 fibers
    #
    if 'path' in kwargs:
        platelistpath = os.path.join(kwargs['path'],'platelist.fits')
    else:
        platelistpath = os.path.join(os.getenv('BOSS_SPECTRO_REDUX'),'platelist.fits')
    platelist = pyfits.open(platelistpath)
    platentotal = platelist[1].data.field('N_TOTAL')
    plateplate = platelist[1].data.field('PLATE')
    platemjd = platelist[1].data.field('MJD')
    platerun2d = platelist[1].data.field('RUN2D')
    platerun1d = platelist[1].data.field('RUN1D')
    platelist.close()
    if 'run2d' in kwargs:
        run2d = kwargs['run2d']
    else:
        run2d = os.getenv('RUN2D')
    if 'run1d' in kwargs:
        run1d = kwargs['run1d']
    else:
        run1d = os.getenv('RUN1D')
    for k in range(mjd.size):
        nfiber[k] = platentotal[(plateplate==plate[k]) & (platemjd==mjd[k]) &
        (platerun2d==run2d) & (platerun1d==run1d)]
    return nfiber



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
def uniq(x,index=None):
    """Replicates the IDL UNIQ() function.

    Returns the *subscripts* of the unique elements of an array.  The elements
    must actually be *sorted* before being passed to this function.  This can
    be done by sorting `x` explicitly or by passing the array subscripts that
    sort `x` as a second parameter.

    Parameters
    ----------
    x : array_like
        Search this array for unique items.
    index : array_like, optional
        This array provides the array subscripts that sort `x`.

    Returns
    -------
    uniq : array_like
        The subscripts of `x` that are the unique elements of `x`.

    Notes
    -----
    Given a sorted array, and assuming that there is a set of
    adjacent identical items, ``uniq()`` will return the subscript of the *last*
    unique item.  This charming feature is retained for reproducibility.

    References
    ----------
    http://www.exelisvis.com/docs/UNIQ.html

    Examples
    --------
    >>> import numpy as np
    >>> from pydl import uniq
    >>> uniq(np.sort(np.array([ 1, 2, 3, 1, 5, 6, 1, 7, 3, 2, 5, 9, 11, 1 ])))
    array([ 3,  5,  7,  9, 10, 11, 12, 13])
    """
    from numpy import array, roll
    if index is None:
        indicies = (x != roll(x,-1)).nonzero()[0]
        if indicies.size > 0:
            return indicies
        else:
            return array([x.size - 1,])
    else:
        q = x[index]
        indicies = (q != roll(q,-1)).nonzero()[0]
        if indicies.size > 0:
            return index[indicies]
        else:
            return array([q.size - 1,],dtype=index.dtype)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def is_cap_used(use_caps,i):
    """Returns True if a cap is used.

    Parameters
    ----------
    use_caps :
    i : int

    Returns
    -------
    is_cap_used : bool
    """
    return (use_caps & 1 << i) != 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Pre-ez_setup bootstrap module to ensure that either distribute or setuptools >=
0.7 is used (over pre-distribute setuptools) if it is available on the path;
otherwise the latest setuptools will be downloaded and bootstrapped with
``ez_setup.py``.
"""

import sys
import imp

try:
    import pkg_resources
    _setuptools_req = pkg_resources.Requirement.parse('setuptools>=0.7')
    # This may raise a DistributionNotFound in which case no version of
    # setuptools or distribute is properly instlaled
    _setuptools = pkg_resources.get_distribution('setuptools')
    if _setuptools not in _setuptools_req:
        # Older version of setuptools; check if we have distribute; again if
        # this results in DistributionNotFound we want to give up
        _distribute = pkg_resources.get_distribution('distribute')
        if _setuptools != _distribute:
            # It's possible on some pathological systems to have an old version
            # of setuptools and distribute on sys.path simultaneously; make
            # sure distribute is the one that's used
            sys.path.insert(1, _distribute.location)
            _distribute.activate()
            imp.reload(pkg_resources)
except:
    # There are several types of exceptions that can occur here; if all else
    # fails bootstrap and use the bootstrapped version
    from ez_setup import use_setuptools
    use_setuptools()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def cirrange(ang,radians=False):
    """Convert an angle larger than 360 degrees to one less than 360 degrees.

    Parameters
    ----------
    ang : float or array_like
        Angle to convert.  If the angle is in radians, the `radians` argument should be set.
    radians : bool
        If true, the input angle is in radians, and the output will be between
        zero and 2*pi.

    Returns
    -------
    cirrange : float or array_like
        Angle in the restricted range.

    Examples
    --------
    >>> from pydl.goddard.misc import cirrange
    >>> cirrange(-270.0)
    90.0
    """
    from math import pi
    if radians:
        cnst = pi * 2.0
    else:
        cnst = 360.0
    #
    # The modulo operator automatically deals with negative values
    #
    return ang % cnst



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def djs_maskinterp1(yval,mask,xval=None,const=False):
    """Interpolate over a masked, 1-d array.

    Parameters
    ----------
    yval : ndarray
        The input values
    mask : ndarray
        The mask
    xval : ndarray, optional
        If set, use these x values, otherwise use an array
    const : bool, optional
        If set to ``True``, bad values around the edges of the array will be
        set to a constant value.

    Returns
    -------
    djs_maskinterp1 : ndarray
    """
    import numpy as np
    good = mask == 0
    if good.all():
        return yval
    ngood = good.sum()
    igood = good.nonzero()[0]
    if ngood == 0:
        return yval
    if ngood == 1:
        return np.zeros(yval.shape,dtype=yval.dtype) + yval[igood[0]]
    ynew = yval.astype('d')
    ny = yval.size
    ibad = (mask != 0).nonzero()[0]
    if xval is None:
        ynew[ibad] = np.interp(ibad,igood,ynew[igood])
        if const:
            if igood[0] != 0:
                ynew[0:igood[0]] = ynew[igood[0]]
            if igood[ngood-1] != ny-1:
                ynew[igood[ngood-1]+1:ny] = ynew[igood[ngood-1]]
    else:
        ii = xval.argsort()
        ibad = (mask[ii] != 0).nonzero()[0]
        igood = (mask[ii] == 0).nonzero()[0]
        ynew[ii[ibad]] = np.interp(xval[ii[ibad]],xval[ii[igood]],ynew[ii[igood]])
        if const:
            if igood[0] != 0:
                ynew[ii[0:igood[0]]] = ynew[ii[igood[0]]]
            if igood[ngood-1] != ny-1:
                ynew[ii[igood[ngood-1]+1:ny]] = ynew[ii[igood[ngood-1]]]
    return ynew
#
#
#
def djs_maskinterp(yval,mask,xval=None,axis=None,const=False):
    """Interpolate over masked pixels in a vector, image or 3-D array.

    Parameters
    ----------
    yval : ndarray
        The input values
    mask : ndarray
        The mask
    xval : ndarray, optional
        If set, use these x values, otherwise use an array
    axis : int, optional
        Must be set if yval has more than one dimension. If set to zero,
        interpolate along the first axis of the array, if set to one,
        interpolate along the second axis of the array, and so on.
    const : bool, optional
        This value is passed to a helper function, djs_maskinterp1.

    Returns
    -------
    djs_maskinterp : ndarray

    """
    import numpy as np
    if mask.shape != yval.shape:
        raise ValueError('mask must have the same shape as yval.')
    if xval is not None:
        if xval.shape != yval.shape:
            raise ValueError('xval must have the same shape as yval.')
    ndim = yval.ndim
    if ndim == 1:
        ynew = djs_maskinterp1(yval,mask,xval,const=const)
    else:
        if axis is None:
            raise ValueError('Must set axis if yval has more than one dimension.')
        if axis < 0 or axis > ndim-1 or axis - int(axis) != 0:
            raise ValueError('Invalid axis value.')
        ynew = np.zeros(yval.shape,dtype=yval.dtype)
        if ndim == 2:
            if xval is None:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        ynew[i,:] = djs_maskinterp1(yval[i,:],mask[i,:],const=const)
                else:
                    for i in range(yval.shape[1]):
                        ynew[:,i] = djs_maskinterp1(yval[:,i],mask[:,i],const=const)
            else:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        ynew[i,:] = djs_maskinterp1(yval[i,:],mask[i,:],xval[i,:],const=const)
                else:
                    for i in range(yval.shape[1]):
                        ynew[:,i] = djs_maskinterp1(yval[:,i],mask[:,i],xval[:,i],const=const)
        elif ndim == 3:
            if xval is None:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[1]):
                            ynew[i,j,:] = djs_maskinterp1(yval[i,j,:],mask[i,j,:],const=const)
                elif axis == 1:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[2]):
                            ynew[i,:,j] = djs_maskinterp1(yval[i,:,j],mask[i,:,j],const=const)
                else:
                    for i in range(yval.shape[1]):
                        for j in range(yval.shape[2]):
                            ynew[:,i,j] = djs_maskinterp1(yval[:,i,j],mask[:,i,j],const=const)
            else:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[1]):
                            ynew[i,j,:] = djs_maskinterp1(yval[i,j,:],mask[i,j,:],xval[i,j,:],const=const)
                elif axis == 1:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[2]):
                            ynew[i,:,j] = djs_maskinterp1(yval[i,:,j],mask[i,:,j],xval[i,j,:],const=const)
                else:
                    for i in range(yval.shape[1]):
                        for j in range(yval.shape[2]):
                            ynew[:,i,j] = djs_maskinterp1(yval[:,i,j],mask[:,i,j],xval[i,j,:],const=const)
        else:
            raise NotImplementedError('Unsupported number of dimensions.')
    return ynew



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def spheregroup(ra,dec,linklength,chunksize=None,debug=False):
    """Perform friends-of-friends grouping given ra/dec coordinates

    Parameters
    ----------
    ra, dec : ndarray
        Arrays of coordinates to group.
    linklength : float
    chunksize : float
    debug : bool

    Returns
    -------
    spheregroup : tuple
    """
    from . import chunks
    from numpy import zeros
    npoints = ra.size
    #
    # Define the chunksize
    #
    if chunksize is not None:
        if chunksize < 4.0*linklength:
            chunksize = 4.0*linklength
            print("chunksize changed to {0:.2f}.".format(chunksize))
    else:
        chunksize = max(4.0*linklength,0.1)
    #
    # Initialize chunks
    #
    chunk = chunks(ra,dec,chunksize)
    chunk.assign(ra,dec,linklength)
    if debug:
        print("raOffset = {0:7.3f}.".format(chunk.raOffset))
        print(chunk.raBounds)
        print(chunk.decBounds)
        print(chunk.chunkList)
    #
    # Run friends-of-friends
    #
    ingroup,multgroup,firstgroup,nextgroup,ngroups = chunk.friendsoffriends(ra,dec,linklength)
    #
    # Renumber the groups in order of appearance
    #
    renumbered = zeros(npoints,dtype='bool')
    iclump = 0
    for i in range(npoints):
        if not renumbered[i]:
            j = firstgroup[ingroup[i]]
            while j != -1:
                ingroup[j] = iclump
                renumbered[j] = True
                j = nextgroup[j]
            iclump += 1
    #
    # Reset the index lists
    #
    firstgroup[:] = -1
    for i in range(npoints-1,-1,-1):
        nextgroup[i] = firstgroup[ingroup[i]]
        firstgroup[ingroup[i]] = i
    #
    # Reset the multiplicities
    #
    for i in range(ngroups):
        j = firstgroup[i]
        while j != -1:
            multgroup[i] += 1
            j = nextgroup[j]
    return (ingroup,multgroup,firstgroup,nextgroup)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_filternum():
    from .. import filternum
    assert filternum('u') == 0
    assert filternum('g') == 1
    assert filternum('r') == 2
    assert filternum('i') == 3
    assert filternum('z') == 4



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the bspline directory of idlutils.
"""
from .bspline import bspline
from .cholesky_band import cholesky_band
from .cholesky_solve import cholesky_solve
from .iterfit import iterfit



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_hogg_iau_name():
    from .. import hogg_iau_name
    assert hogg_iau_name(354.120375,-0.544777778) == 'SDSS J233628.89-003241.2'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def pca_solve(flux,ivar,loglam=None,zfit=None,**kwargs):
    """Replacement for idlspec2d pca_solve.pro
    """
    import time
    import numpy as np
    import pydl.pydlspec2d.spec1d
    import pydl.pydlspec2d.spec2d
    from ... import pcomp
    from ...pydlutils.math import computechi2, djs_reject
    if 'maxiter' in kwargs:
        maxiter = kwargs['maxiter']
    else:
        maxiter = 0
    if 'niter' in kwargs:
        niter = kwargs['niter']
    else:
        niter = 10
    if 'nkeep' in kwargs:
        nkeep = kwargs['nkeep']
    else:
        nkeep = 3
    if 'nreturn' in kwargs:
        nreturn = kwargs['nreturn']
    else:
        nreturn = nkeep
    if len(flux.shape) == 1:
        nobj = 1
        npix = flux.shape[0]
    else:
        nobj, npix = flux.shape
    print("Building PCA from {0:d} object spectra.".format(nobj))
    #
    # The redshift of each object in pixels would be logshift/objdloglam.
    #
    if zfit is None:
        logshift = np.zeros((nobj,),dtype=flux.dtype)
    else:
        logshift = np.log10(1.0 + zfit)
    #
    # Determine the new wavelength mapping.
    #
    if loglam is None:
        newflux = flux
        newivar = ivar
        newloglam = kwargs['newloglam']
        nnew = flux.shape[1]
    else:
        if 'newloglam' in kwargs:
            fullloglam = kwargs['newloglam']
            dloglam = fullloglam[1] - fullloglam[0]
        else:
            igood = loglam != 0
            dloglam = loglam[1] - loglam[0]
            logmin = loglam[igood].min() - logshift.max()
            logmax = loglam[igood].max() - logshift.min()
            if 'wavemin' in kwargs:
                logmin = max(logmin, np.log10(wavemin))
            if 'wavemax' in kwargs:
                logmax = min(logmax,np.log10(wavemax))
            fullloglam = pydl.pydlspec2d.spec1d.wavevector(logmin,logmax,binsz=dloglam)
        nnew = fullloglam.size
        fullflux = np.zeros((nobj,nnew),dtype='d')
        fullivar = np.zeros((nobj,nnew),dtype='d')
        #
        # Shift each spectrum to z = 0 and sample at the output wavelengths
        #
        if loglam.ndim == 1:
            indx = loglam > 0
            rowloglam = loglam[indx]
        for iobj in range(nobj):
            print("OBJECT {0:5d}".format(iobj))
            if loglam.ndim > 1:
                if loglam.shape[0] != nobj:
                    raise ValueError('Wrong number of dimensions for loglam.')
                indx = loglam[iobj,:] > 0
                rowloglam = loglam[iobj,indx]
            flux1,ivar1 = pydl.pydlspec2d.spec2d.combine1fiber(rowloglam-logshift[iobj],flux[iobj,indx],
                ivar[iobj,indx],newloglam=fullloglam,binsz=dloglam,aesthetics='mean') # ,verbose=True)
            fullflux[iobj,:] = flux1
            fullivar[iobj,:] = ivar1
        #
        # Find the columns out side of which there is no data at all
        #
        # nzi = fullivar.nonzero()
        # firstcol = nzi[1].min()
        # lastcol = nzi[1].max()
        # newflux = fullflux[:,firstcol:lastcol+1]
        # newivar = fullivar[:,firstcol:lastcol+1]
        # newloglam = fullloglam[firstcol:lastcol+1]
        # nnew = newloglam.size
        newflux = fullflux
        newivar = fullivar
        newloglam = fullloglam
    nzi = newivar.nonzero()
    first_nonzero = (np.arange(nobj,dtype=nzi[0].dtype),
        np.array([nzi[1][nzi[0]==k].min() for k in range(nobj)]))
    #
    # Construct the synthetic weight vector, to be used when replacing the
    # low-S/N object pixels with the reconstructions.
    #
    synwvec = np.ones((nnew,),dtype='d')
    for ipix in range(nnew):
        indx = newivar[:,ipix] != 0
        if indx.any():
            synwvec[ipix] = newivar[indx,ipix].mean()
    fluxdict = {'flux':newflux, 'eigenval':1.0, 'acoeff':1.0,
        'outmask':np.ones((nnew,),dtype='i4'),
        'usemask':np.ones((nnew,),dtype='i4'),
        'newflux':newflux,'newivar':newivar,
        'newloglam':newloglam, }
        # 'emevecs':1.0, 'emevals':1.0}
    #
    # If there is only one object spectrum, then all we can do is return it.
    #
    if nobj == 1:
        fluxdict['flux'] = newflux.astype('f')
        return fluxdict
    #
    # Rejection iteration loop.
    #
    qdone = 0
    iiter = 0
    #
    # Begin with all points good.
    #
    outmask = None
    inmask = newivar != 0
    ymodel = None
    # emevecs, emevals = pydlutils.empca(newflux,inmask)
    # fluxdict['emevecs'] = emevecs
    # fluxdict['emevals'] = emeveals
    while qdone == 0 and iiter <= maxiter:
        # print 'starting djs_reject'
        outmask,qdone = djs_reject(newflux,ymodel,
            inmask=inmask,outmask=outmask,
            invvar=newivar)
        # print 'finished with djs_reject'
        #
        # Iteratively do the PCA solution
        #
        filtflux = newflux.copy()
        acoeff = np.zeros((nobj,nkeep),dtype='d')
        t0 = time.time()
        for ipiter in range(niter):
            #
            # We want to get these values from the pcomp routine.
            #
            # eigenval = 1
            # coeff = 1
            flux0 = np.tile(filtflux[first_nonzero],nnew).reshape(nnew,nobj).transpose()
            # flux0 = np.tile(filtflux,nnew).reshape(nnew,nobj).transpose()
            totflux = np.absolute(filtflux - flux0).sum(1)
            goodobj = totflux > 0
            if goodobj.all():
                tmp = pcomp(filtflux.T) # ,standardize=True)
                pres = tmp.derived
                eigenval = tmp.eigenvalues
            else:
                tmp = pcomp(filtflux[goodobj,:].T) # ,standardize=True)
                pres = np.zeros((nobj,nnew),dtype='d')
                pres[goodobj,:] = tmp.derived
                eigenval = np.zeros((nobj,),dtype='d')
                eigenval[goodobj] = tmp.eigenvalues
            maskivar = newivar * outmask
            sqivar = np.sqrt(maskivar)
            for iobj in range(nobj):
                out = computechi2(newflux[iobj,:], sqivar[iobj,:],
                    pres[:,0:nkeep])
                filtflux[iobj,:] = (maskivar[iobj,:] * newflux[iobj,:] +
                    synwvec*out.yfit) / (maskivar[iobj,:] + synwvec)
                acoeff[iobj,:] = out.acoeff
            if 'quiet' not in kwargs:
                print("The elapsed time for iteration #{0:2d} is {1:6.2f} s.".format(ipiter+1,time.time()-t0))
        #
        # Now set ymodel for rejecting points.
        #
        ymodel = np.dot(acoeff,pres[:,0:nkeep].T)
        iiter += 1
    if nobj == 1:
        usemask = outmask
    else:
        usemask = outmask.sum(0)
    fluxdict['usemask'] = usemask
    fluxdict['outmask'] = outmask
    fluxdict['flux'] = pres[:,0:nreturn].transpose().astype('f')
    fluxdict['eigenval'] = eigenval[0:nreturn]
    fluxdict['acoeff'] = acoeff
    return fluxdict




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
class groups(object):
    """Group a set of objects (a list of coordinates in some space) based on
    a friends-of-friends algorithm
    """
    import numpy as np
    from .. import PydlutilsException
    @staticmethod
    def euclid(x1,x2):
        """Pythagorean theorem in Euclidean space with arbitrary number
        of dimensions.
        """
        return self.np.sqrt(((x1-x2)**2).sum())
    @staticmethod
    def sphereradec(x1,x2):
        """Separation of two points on a 2D-sphere, assuming they are in
        longitude-latitude or right ascension-declination form.  Assumes
        everything is already in radians.
        """
        from ...goddard.astro import gcirc
        return gcirc(x1[0],x1[1],x2[0],x2[1],units=0)
    def __init__(self,coordinates,distance,separation='euclid'):
        """Init creates an object and performs the friends-of-friends
        algorithm.  The coordinates can have arbitrary dimensions, with each
        column representing one of the dimensions.  Each row defines an object.
        If separation is not defined it defaults to Euclidean space.
        """
        #
        # Find a separation function
        #
        if callable(separation):
            self.separation = separation
        elif isinstance(separation,str):
            if separation == 'euclid':
                self.separation = self.euclid
            elif separation == 'sphereradec':
                self.separation = self.sphereradec
            else:
                raise self.PydlutilsException("Unknown separation function: {0}.".format(separation))
        else:
            raise self.PydlutilsException("Improper type for separation!")
        #
        # Save information about the coordinates.
        #
        nGroups = 0
        nTargets = coordinates.shape[0]
        multGroup = self.np.zeros(nTargets,dtype='i4')
        firstGroup = self.np.zeros(nTargets,dtype='i4') -1
        nextGroup = self.np.zeros(nTargets,dtype='i4') -1
        inGroup = self.np.arange(nTargets,dtype='i4')
        #
        # Find all the other targets associated with each target
        #
        for i in range(nTargets):
            nTmp = 0
            minGroup = nGroups
            for j in range(nTargets):
                sep = self.separation(coordinates[i,:],coordinates[j,:])
                if sep <= distance:
                    multGroup[nTmp] = j
                    minGroup = min(minGroup,inGroup[j])
                    nTmp += 1
            #
            # Use this minimum for all
            #
            for j in range(nTmp):
                if inGroup[multGroup[j]] < nTargets:
                    k = firstGroup[inGroup[multGroup[j]]]
                    while k != -1:
                        inGroup[k] = minGroup
                        k = nextGroup[k]
                inGroup[multGroup[j]] = minGroup
            #
            # If it is a new group (no earlier groups), increment nGroups
            #
            if minGroup == nGroups:
                nGroups += 1
            for j in range(i+1):
                firstGroup[j] = -1
            for j in range(i,-1,-1):
                nextGroup[j] = firstGroup[inGroup[j]]
                firstGroup[inGroup[j]] = j
        #
        # Renumber to get rid of the numbers which were skipped
        #
        renumber = self.np.zeros(nTargets,dtype='bool')
        nTmp = nGroups
        nGroups = 0
        for i in range(nTargets):
            if not renumbered[i]:
                j = firstGroup[inGroup[i]]
                while j != -1:
                    inGroup[j] = nGroups
                    renumbered[j] = True
                    j = nextGroup[j]
                nGroups += 1
        #
        # Reset the values of firstGroup and inGroup
        #
        firstGroup[:] = -1
        for i in range(nTargets-1,-1,-1):
            nextGroup[i] = firstGroup[inGroup[i]]
            firstGroup[inGroup[i]] = i
        #
        # Get the multiplicity
        #
        for i in range(nGroups):
            multGroup[i] = 0
            j = firstGroup[i]
            while j != -1:
                multGroup[i] += 1
                j = nextGroup[j]
        #
        # Set attributes
        #
        self.nGroups = nGroups
        self.nTargets = nTargets
        self.inGroup = inGroup
        self.multGroup = multGroup
        self.firstGroup = firstGroup
        self.nextGroup = nextGroup
        return
    def inGroup(self,i):
        return self.inGroup[i]
    def multGroup(self,i):
        return self.multGroup[i]
    def firstGroup(self,i):
        return self.firstGroup[i]
    def nextGroup(self,i):
        return self.nextGroup[i]




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def plot_eig(filename,title='Unknown'):
    """Plot spectra from an eigenspectra/template file.

    Parameters
    ----------
    filename : str
        Name of a FITS file containing eigenspectra/templates.
    title : str, optional
        Title to put on the plot.

    Returns
    -------
    None
    """
    from astropy.io import fits as pyfits
    import pylab
    #
    # Set title based on filename
    #
    if title == 'Unknown':
        if filename.find('Gal') > 0:
            title = 'Galaxies: Eigenspectra'
        elif filename.find('QSO') > 0:
            title = 'QSOs: Eigenspectra'
        elif filename.find('Star') > 0:
            title = 'Stars: Eigenspectra'
        elif filename.find('CVstar') > 0:
            title = 'CV Stars: Eigenspectra'
        else:
            print('Unknown template type!')
    base,ext = filename.split('.')
    spectrum = pyfits.open(filename)
    newloglam0 = spectrum[0].header['COEFF0']
    objdloglam = spectrum[0].header['COEFF1']
    spectro_data = spectrum[0].data
    spectrum.close()
    (neig, ndata) = spectro_data.shape
    newloglam = pylab.arange(ndata) * objdloglam + newloglam0
    lam = 10.0**newloglam
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    colorvec = ['k','r','g','b','m','c']
    for l in range(neig):
        p = ax.plot(lam,spectro_data[l,:],'%s-'%colorvec[l%len(colorvec)],linewidth=1)
    ax.set_xlabel(r'Wavelength [$\AA$]')
    ax.set_ylabel('Flux [Arbitrary Units]')
    ax.set_title(title)
    # ax.set_xlim([3500.0,10000.0])
    # ax.set_ylim([-400.0,500.0])
    # fig.savefig(base+'.zoom.png')
    fig.savefig(base+'.png')
    pylab.close(fig)
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Create galaxy template files.
"""
from __future__ import print_function
#
def pca_gal(**kwargs):
    """Wrapper on pca_solve to handle galaxy eigenspectra.

    Keyword Parameters
    ------------------
    inputfile : str, optional
        The list of spectra to use.  If not specified, $IDLSPEC2D_DIR/tempates/eigeninput_gal.dat will be used.
    wavemin : float, optional
        Minimum wavelength for the template.  If not specified 1850 Å will be used.
    wavemax : float, optional
        Maximum wavelength for the template.  If not specified 10000 Å will be used.
    niter : int, optional
        Number of iterations.  The default is 10.
    dump : str, optional
        If set, save input data in a Python pickle file.
    flux : bool, optional
        If set to ``True`` make some additional QA plots of the input spectra.
        The default is ``False``.

    Returns
    -------
    None

    Side Effects
    ------------
    Creates spEigenGal-MJD.fits and some associated QA plots.
    """
    import os
    import os.path
    import pickle
    import matplotlib
    matplotlib.use('Agg') # Non-interactive back-end
    import pylab
    from astropy.io import fits as pyfits
    import numpy as np
    from matplotlib.font_manager import fontManager, FontProperties
    from ...goddard.astro import get_juldate
    from ...pydlutils.image import djs_maskinterp
    from ...pydlutils.math import djs_median
    #from ...pydlutils.misc import djs_readcol
    from astropy.io import ascii
    from . import pca_solve, plot_eig, readspec, skymask, wavevector
    if 'inputfile' in kwargs:
        inputfile = kwargs['inputfile']
    else:
        inputfile = os.path.join(os.getenv('IDLSPEC2D_DIR'),
            'templates','eigeninput_gal.dat')
    if 'wavemin' in kwargs:
        wavemin = kwargs['wavemin']
    else:
        wavemin = 1850.0
    if 'wavemax' in kwargs:
        wavemax = kwargs['wavemax']
    else:
        wavemax = 10000.0
    snmax = 100.0
    if 'niter' in kwargs:
        niter = kwargs['niter']
    else:
        niter = 10
    nkeep = 4
    minuse = 10
    #
    # Name the output files.
    #
    jd = get_juldate()
    outfile = "spEigenGal-%d" % int(jd - 2400000.5)
    #
    # Read the input spectra
    #
    #plate,mjd,fiber,zfit = djs_readcol(inputfile,format='(L,L,L,D)')
    converters = {'plate': [ascii.convert_numpy(np.int32)],
        'mjd': [ascii.convert_numpy(np.int32)],
        'fiber': [ascii.convert_numpy(np.int32)] }
    input_data = ascii.read(inputfile,names=['plate','mjd','fiber','zfit'],converters=converters)
    plate = input_data['plate'].data
    mjd = input_data['mjd'].data
    fiber = input_data['fiber'].data
    zfit = input_data['zfit'].data
    if 'dump' in kwargs:
        dumpfile = kwargs['dump']
    else:
        dumpfile = 'this-file-does-not-exist'
    if os.path.exists(dumpfile):
        print("Loading data from {0}.".format(dumpfile))
        f = open(dumpfile)
        pcaflux = pickle.load(f)
        f.close()
    else:
        spplate = readspec(plate,fiber,mjd=mjd,**kwargs)
        #
        # Insist that all of the requested spectra exist.
        #
        missing = spplate['plugmap']['FIBERID'] == 0
        if missing.any():
            imissing = missing.nonzero()[0]
            for k in imissing:
                print("Missing plate={0:d} mjd={1:d} fiber={2:d}".format(plate[k],mjd[k],fiber[k]))
            raise ValueError("{0:d} missing object(s).".format(missing.sum()))
        #
        # Do not fit where the spectrum may be dominated by sky-sub residuals.
        #
        objinvvar = skymask(spplate['invvar'],spplate['andmask'],spplate['ormask'])
        ifix = spplate['flux']**2 * objinvvar > snmax**2
        if ifix.any():
            objinvvar[ifix.nonzero()] = (snmax/spplate['flux'][ifix.nonzero()])**2
        #
        # Set the new wavelength mapping here.  If the binsz keyword is not set,
        # then bin size is determined from the first spectrum returned by readspec.
        # This is fine in the case where all spectra have the same bin size
        # (though their starting wavelengths may differ).  However, this may not
        # be a safe assumption in the future.
        #
        if 'binsz' in kwargs:
            objdloglam = kwargs['binsz']
        else:
            objdloglam = spplate['loglam'][0,1] - spplate['loglam'][0,0]
        newloglam = wavevector(np.log10(wavemin),np.log10(wavemax),
            binsz=objdloglam)
        #
        # Do PCA solution.
        #
        pcaflux = pca_solve(spplate['flux'],objinvvar,spplate['loglam'],zfit,
            niter=niter,nkeep=nkeep,newloglam=newloglam,aesthetics='mean')
        #
        # Fill in bad data with a running median of the good data.
        #
        qgood = pcaflux['usemask'] >= minuse
        medflux = np.zeros(pcaflux['flux'].shape,dtype=pcaflux['flux'].dtype)
        if not qgood.all():
            for i in range(nkeep):
                medflux[i,qgood] = djs_median(pcaflux['flux'][i,qgood],
                    width=51,boundary='nearest')
                medflux[i,:] = djs_maskinterp(medflux[i,:],~qgood,const=True)
            pcaflux['flux'][:,~qgood] = medflux[:,~qgood]
        #
        # Dump input fluxes to a file for debugging purposes.
        #
        if 'dump' in kwargs:
            f = open(kwargs['dump'],'w')
            pickle.dump(pcaflux, f)
            f.close()
    #
    # Make plots
    #
    colorvec = ['k','r','g','b','m','c']
    smallfont = FontProperties(size='xx-small');
    nspectra = pcaflux['newflux'].shape[0]
    if 'flux' in kwargs:
        nfluxes = 30
        separation = 5.0
        nplots = nspectra/nfluxes
        if nspectra % nfluxes > 0:
            nplots += 1
        for k in range(nplots):
            istart = k*nfluxes
            iend = min(istart+nfluxes,nspectra) - 1
            fig = pylab.figure(dpi=100)
            ax = fig.add_subplot(111)
            for l in range(istart,iend+1):
                p = ax.plot(10.0**pcaflux['newloglam'],pcaflux['newflux'][l,:]+separation*(l%nfluxes),
                    '%s-'%colorvec[l%len(colorvec)],linewidth=1)
            ax.set_xlabel(r'Wavelength [$\AA$]')
            ax.set_ylabel(r'Flux [$\mathsf{10^{-17} erg\, cm^{-2} s^{-1} \AA^{-1}}$] + Constant')
            ax.set_title('Galaxies: Input Spectra %4d-%4d' % (istart+1,iend+1))
            ax.set_ylim(pcaflux['newflux'][istart,:].min(),pcaflux['newflux'][iend-1,:].max()+separation*(nfluxes-1))
            fig.savefig('%s.flux.%04d-%04d.png'%(outfile,istart+1,iend+1))
            pylab.close(fig)
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(10.0**pcaflux['newloglam'],(pcaflux['newivar'] == 0).sum(0)/float(nspectra),'k-')
    ax.set_xlabel(r'Wavelength [$\AA$]')
    ax.set_ylabel('Fraction of spectra with missing data')
    ax.set_title('Missing Data')
    fig.savefig(outfile+'.missing.png')
    pylab.close(fig)
    aratio10 = pcaflux['acoeff'][:,1]/pcaflux['acoeff'][:,0]
    aratio20 = pcaflux['acoeff'][:,2]/pcaflux['acoeff'][:,0]
    aratio30 = pcaflux['acoeff'][:,3]/pcaflux['acoeff'][:,0]
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(aratio10,aratio20,marker='None',linestyle='None')
    for k in range(len(aratio10)):
        t = ax.text(aratio10[k],aratio20[k],'%04d-%04d'%(plate[k],fiber[k]),
            horizontalalignment='center', verticalalignment='center',
            color=colorvec[k%len(colorvec)],
            fontproperties=smallfont)
    # ax.set_xlim([aratio10.min(),aratio10.max])
    # ax.set_xlim([aratio20.min(),aratio20.max])
    ax.set_xlabel('Eigenvalue Ratio, $a_1/a_0$')
    ax.set_ylabel('Eigenvalue Ratio, $a_2/a_0$')
    ax.set_title('Galaxies: Eigenvalue Ratios')
    fig.savefig(outfile+'.a2_v_a1.png')
    pylab.close(fig)
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(aratio20,aratio30,marker='None',linestyle='None')
    for k in range(len(aratio10)):
        t = ax.text(aratio20[k],aratio30[k],'%04d-%04d'%(plate[k],fiber[k]),
            horizontalalignment='center', verticalalignment='center',
            color=colorvec[k%len(colorvec)],
            fontproperties=smallfont)
    # ax.set_xlim([aratio10.min(),aratio10.max])
    # ax.set_xlim([aratio20.min(),aratio20.max])
    ax.set_xlabel('Eigenvalue Ratio, $a_2/a_0$')
    ax.set_ylabel('Eigenvalue Ratio, $a_3/a_0$')
    ax.set_title('Galaxies: Eigenvalue Ratios')
    fig.savefig(outfile+'.a3_v_a2.png')
    pylab.close(fig)
    #
    # Save output to FITS file.
    #
    if os.path.exists(outfile+'.fits'):
        os.remove(outfile+'.fits')
    hdu0 = pyfits.PrimaryHDU(pcaflux['flux'])
    hdu1 = pyfits.new_table(pyfits.ColDefs([
        pyfits.Column(name='plate',format='J',array=plate),
        pyfits.Column(name='mjd',format='J',array=mjd),
        pyfits.Column(name='fiber',format='J',array=fiber),
        pyfits.Column(name='redshift',format='D',array=zfit)]))
    hdulist = pyfits.HDUList([hdu0,hdu1])
    hdulist[0].header.update('OBJECT','GALAXY')
    hdulist[0].header.update('COEFF0',pcaflux['newloglam'][0])
    hdulist[0].header.update('COEFF1',pcaflux['newloglam'][1]-pcaflux['newloglam'][0])
    hdulist[0].header.update('IDLUTILS','pydlutils','Version of idlutils')
    hdulist[0].header.update('SPEC2D','eigenspectra','Version of idlspec2d')
    hdulist[0].header.update('RUN2D',os.getenv('RUN2D'),'Version of 2d reduction')
    hdulist[0].header.update('RUN1D',os.getenv('RUN1D'),'Version of 1d reduction')
    for i in range(len(pcaflux['eigenval'])):
        hdulist[0].header.update("EIGEN%d" % i,pcaflux['eigenval'][i])
    hdulist[1].header.update('FILENAME',inputfile)
    hdulist.writeto(outfile+'.fits')
    plot_eig(outfile+'.fits')
    return
#
#
#
def main():
    """Gets called when the script is executed.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Side-Effects
    ------------
    Creates a FITS file and some PNG plots.
    """
    #
    # Imports for main()
    #
    import os
    import os.path
    import sys
    from astropy.utils.compat import argparse
    #
    # Get Options
    #
    parser = argparse.ArgumentParser(description=__doc__,prog=os.path.basename(sys.argv[0]))
    parser.add_argument('-F', '--flux', action='store_true', dest='flux',
        help='Plot input spectra.')
    parser.add_argument('-d', '--dump', action='store', dest='dump',
        metavar='FILE', default=os.path.join(os.getenv('HOME'),'scratch','eigeninput_gal.dump'),
        help='Dump data to a pickle file.')
    #parser.add_argument('-n', '--nonnegative', action='store_true', dest='nonnegative',
    #    help='Use non-negative HMF method.')
    #parser.add_argument('-e', '--epsilon', action='store', type=float, dest='epsilon',
    #    metavar='EPSILON', default=1.0, help='Set the epsilon parameter (default 1.0). Set to 0 to turn off entirely')
    #parser.add_argument('-K', '--dims', action='store', type=int, dest='K',
    #    metavar='K', default=4, help='Set the number of functions to model (default 4).')
    #parser.add_argument('-d', '--outdir', action='store', dest='outdir',
    #    metavar='DIR', default=os.path.join(os.getenv('HOME'),'scratch'),
    #    help='Write output files to DIR.')
    parser.add_argument('-f', '--file', action='store', dest='inputfile',
        metavar='FILE', default=os.path.join(os.getenv('HOME'),'scratch','eigeninput_gal.dat'),
        help='Read input spectra and redshifts from FILE.')
    options = parser.parse_args()
    pca_gal(inputfile=options.inputfile,dump=options.dump,
        flux=options.flux)
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def findspec(*args,**kwargs):
    """Find SDSS/BOSS spectra that match a given RA, Dec.

    Parameters
    ----------

    Returns
    -------
    """
    import os
    import os.path
    import glob
    from astropy.io import fits as pyfits
    import numpy as np
    from ... import uniq
    from ...pydlutils.misc import djs_readcol, struct_print
    from ...pydlutils.spheregroup import spherematch
    from .. import Pydlspec2dException
    import pydl.pydlspec2d.spec1d # get the findspec_cache dictionary
    #
    # Set up default values
    #
    if 'sdss' in kwargs:
        topdir = os.getenv('SPECTRO_REDUX')
        run2d = '26'
        run1d = ''
    else:
        if 'topdir' in kwargs:
            topdir = kwargs['topdir']
        else:
            topdir = os.getenv('BOSS_SPECTRO_REDUX')
        if 'run2d' in kwargs:
            run2d = str(kwargs['run2d'])
        else:
            run2d = os.getenv('RUN2D')
        if 'run1d' in kwargs:
            run1d = str(kwargs['run1d'])
        else:
            run1d = os.getenv('RUN1D')
    if pydl.pydlspec2d.spec1d.findspec_cache is None:
        pydl.pydlspec2d.spec1d.findspec_cache = { 'lasttopdir':topdir,
            'plist':None }
    if (pydl.pydlspec2d.spec1d.findspec_cache['plist'] is None or
        topdir != pydl.pydlspec2d.spec1d.findspec_cache['lasttopdir']):
        pydl.pydlspec2d.spec1d.findspec_cache['lasttopdir'] = topdir
        platelist_file = os.path.join(topdir,"platelist.fits")
        plates_files = glob.glob(os.path.join(topdir,"plates-*.fits"))
        plist = None
        if os.path.exists(platelist_file):
            platelist = pyfits.open(platelist_file)
            plist = platelist[1].data
            platelist.close()
        if len(plates_files) > 0:
            plates = pyfits.open(plates_files[0])
            plist = plates[1].data
            plates.close()
        if plist is None:
            raise Pydlspec2dException("Plate list (platelist.fits or plates-*.fits) not found in {0}.".format(topdir))
        else:
            pydl.pydlspec2d.spec1d.findspec_cache['plist'] = plist
    qdone = plist.field('STATUS1D') == 'Done'
    qdone2d = plist.field('RUN2D').strip() == run2d
    if run1d == '':
        qdone1d = np.ones(plist.size,dtype='bool')
    else:
        qdone1d = plist.field('RUN1D').strip() == run1d
    qfinal = qdone & qdone2d & qdone1d
    if not qfinal.any():
        print("No reduced plates!")
        return None
    idone = np.arange(plist.size)[qfinal]
    #
    # If there are positional arguments, interpret these as RA, Dec
    #
    if len(args) == 2:
        ra = args[0]
        dec = args[1]
    #
    # Read RA, Dec from infile if set
    #
    if 'infile' in kwargs:
        ra, dec = djs_readcol(kwargs['infile'],format='(D,D)')
    if 'searchrad' in kwargs:
        searchrad = float(kwargs['searchrad'])
    else:
        searchrad = 3.0/3600.0
    #
    # Create output structure
    #
    slist_type = np.dtype([('PLATE','i4'),('MJD','i4'),('FIBERID','i4'),
        ('RA','f8'),('DEC','f8'),('MATCHRAD','f8')])
    #
    # Match all plates with objects
    #
    imatch1, itmp, dist12 = spherematch(ra, dec,
        plist[qfinal].field('RACEN'), plist[qfinal].field('DECCEN'),
        searchrad+1.55,maxmatch=0)
    if imatch1.size == 0:
        return None
    imatch2 = idone[itmp]
    #
    # Read all relevant plates
    #
    try:
        n_total = plist.field('N_TOTAL')
    except KeyError:
        n_total = np.zeros(plist.size,dtype='i4') + 640
    iplate = imatch2[uniq(imatch2,imatch2.argsort())]
    i0 = 0
    plugmap = np.zeros(n_total[iplate].sum(),
        dtype=[('PLATE','i4'),('MJD','i4'),('FIBERID','i4'),
        ('RA','d'),('DEC','d')])
    for i in range(iplate.size):
        spplate = pydl.pydlspec2d.spec1d.readspec(plist[iplate[i]].field('PLATE'),mjd=plist[iplate[i]].field('MJD'),
            topdir=topdir,run2d=run2d,run1d=run1d)
        index_to = i0 + np.arange(n_total[iplate[i]],dtype='i4')
        plugmap['PLATE'][index_to] = plist[iplate[i]].field('PLATE')
        plugmap['MJD'][index_to] = plist[iplate[i]].field('MJD')
        plugmap['FIBERID'][index_to] = spplate['plugmap']['FIBERID']
        plugmap['RA'][index_to] = spplate['plugmap']['RA']
        plugmap['DEC'][index_to] = spplate['plugmap']['DEC']
        i0 += n_total[iplate[i]]
    i1, i2, d12 = spherematch(ra, dec, plugmap['RA'], plugmap['DEC'], searchrad, maxmatch=0)
    if i1.size == 0:
        return None
    if 'best' in kwargs:
        #
        # Return only best match per object
        #
        slist = np.zeros(ra.size,dtype=slist_type)
        spplate = pydl.pydlspec2d.spec1d.readspec(plugmap[i2]['PLATE'],plugmap[i2]['FIBERID'],
            mjd=plugmap[i2]['MJD'],topdir=topdir, run2d=run2d, run1d=run1d)
        sn = spplate['zans']['SN_MEDIAN']
        isort = (i1 + np.where(sn > 0, sn, 0)/(sn+1.0).max()).argsort()
        i1 = i1[isort]
        i2 = i2[isort]
        d12 = d12[isort]
        iuniq = uniq(i1)
        slist[i1[iuniq]]['PLATE'] = plugmap[i2[iuniq]]['PLATE']
        slist[i1[iuniq]]['MJD'] = plugmap[i2[iuniq]]['MJD']
        slist[i1[iuniq]]['FIBERID'] = plugmap[i2[iuniq]]['FIBERID']
        slist[i1[iuniq]]['RA'] = plugmap[i2[iuniq]]['RA']
        slist[i1[iuniq]]['DEC'] = plugmap[i2[iuniq]]['DEC']
        slist[i1[iuniq]]['MATCHRAD'] = d12[iuniq]
    else:
        #
        # Return all matches
        #
        slist = np.zeros(i1.size,dtype=slist_type)
        slist['PLATE'] = plugmap[i2]['PLATE']
        slist['MJD'] = plugmap[i2]['MJD']
        slist['FIBERID'] = plugmap[i2]['FIBERID']
        slist['RA'] = plugmap[i2]['RA']
        slist['DEC'] = plugmap[i2]['DEC']
        slist['MATCHRAD'] = d12
    #
    # Print to terminal or output file
    #
    if 'print' in kwargs:
        foo = struct_print(slist)
    if 'outfile' in kwargs:
        foo = struct_print(slist,filename=outfile)
    return slist



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_flagexist(flagname, bitname, flagexist=False, whichexist=False):
    """Check for the existence of flags.

    Parameters
    ----------
    flagname : str
        The name of a bitmask group. Not case-sensitive.
    bitname : str or list
        The name(s) of the specific bitmask(s) within the `flagname` group.
    flagexist : bool, optional
        If flagexist is True, return a tuple with the second component indicating
        whether the binary flag named `flagname` exists, even if `bitname` is wrong.
    whichexist : bool, optional
        If whichexist is True, return a list containing existence test results
        for each individual flag.

    Returns
    -------
    sdss_flagexist : bool or tuple
        A boolean value or a tuple of bool.
    """
    from . import maskbits
    if maskbits is None:
        from .set_maskbits import set_maskbits
        maskbits = set_maskbits()
    #
    # Make sure label is a list
    #
    if isinstance(bitname,str):
        bitnames = [bitname.upper()]
    else:
        bitnames = [b.upper() for b in bitname]
    f = False
    l = False
    which = [False]*len(bitnames)
    if flagname.upper() in maskbits:
        f = True
        which = [n in maskbits[flagname.upper()] for n in bitnames]
        l = sum(which) == len(which)
    if flagexist and whichexist:
        return (l,f,which)
    elif flagexist:
        return (l,f)
    elif whichexist:
        return (l,which)
    else:
        return l



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This is the pydl/pydlutils/yanny/tests directory.
"""

import os
import shutil
import stat
import tempfile
import time
import warnings

from ... import fits


class YannyTestCase(object):
    """Based on astropy.io.fits.tests.FitsTestCase."""
    save_temp = False
    def setup(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), 't')
        self.temp_dir = tempfile.mkdtemp(prefix='yanny-test-')
        # Ignore deprecation warnings--this only affects Python 2.5 and 2.6,
        # since deprecation warnings are ignored by defualt on 2.7
        warnings.simplefilter('ignore')
        warnings.simplefilter('always', UserWarning)

    def teardown(self):
        warnings.resetwarnings()
        if not self.save_temp:
            if hasattr(self, 'temp_dir') and os.path.exists(self.temp_dir):
                tries = 3
                while tries:
                    try:
                        shutil.rmtree(self.temp_dir)
                        break
                    except OSError:
                        # Probably couldn't delete the file because for whatever
                        # reason a handle to it is still open/hasn't been
                        # garbage-collected
                        time.sleep(0.5)
                        tries -= 1

    def copy_file(self, filename):
        """Copies a backup of a test data file to the temp dir and sets its
        mode to writeable.
        """

        shutil.copy(self.data(filename), self.temp(filename))
        os.chmod(self.temp(filename), stat.S_IREAD | stat.S_IWRITE)

    def data(self, filename):
        """Returns the path to a test data file."""

        return os.path.join(self.data_dir, filename)

    def temp(self, filename):
        """ Returns the full path to a file in the test temp dir."""

        return os.path.join(self.temp_dir, filename)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the spheregroup directory of idlutils.
"""
from .chunks import chunks
from .groups import groups
from .spheregroup import spheregroup
from .spherematch import spherematch



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the sdss directory of idlutils.
"""
from .default_skyversion import default_skyversion
from .sdss_astrombad import sdss_astrombad
from .sdss_flagexist import sdss_flagexist
from .sdss_flagname import sdss_flagname
from .sdss_flagval import sdss_flagval
from .sdss_objid import sdss_objid
from .sdss_sweep_circle import sdss_sweep_circle
#
# Cache for the maskbits file.
#
maskbits = None
#
# Cache the sweep index
#
sweep_cache = {'star':None,'gal':None,'sky':None}
#
# Cache sdss_astrombad data
#
opbadfields = None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the mangle directory of idlutils.
"""
from .is_cap_used import is_cap_used
from .read_fits_polygons import read_fits_polygons
from .set_use_caps import set_use_caps



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_objid():
    from .. import sdss_objid
    assert sdss_objid(3704,3,91,146) == 1237661382772195474



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def combine1fiber(inloglam,objflux,objivar=None,**kwargs):
    """Combine several spectra of the same object, or resample a single spectrum.
    """
    import numpy as np
    import pydl.pydlutils.bspline
    import pydl.pydlutils.sdss
    from pydl import smooth
    from . import aesthetics
    #
    # Check that dimensions of inputs are valid.
    #
    if 'newloglam' in kwargs:
        newloglam = kwargs['newloglam']
        nfinalpix = len(newloglam)
    else:
        raise ValueError('newloglam is required.')
    if objflux.shape != inloglam.shape:
        raise ValueError('Dimensions of inloglam and objflux do not agree.')
    if objivar is not None:
        if objivar.shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and objivar do not agree.')
    if 'finalmask' in kwargs:
        if kwargs['finalmask'].shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and finalmask do not agree.')
    if 'indisp' in kwargs:
        if kwargs['indisp'].shape != inloglam.shape:
            raise ValueError('Dimensions of inloglam and indisp do not agree.')
    #
    # Set defaults
    #
    EPS = np.finfo(np.float32).eps
    if 'binsz' in kwargs:
        binsz = kwargs['binsz']
    else:
        binsz = inloglam[1] - inloglam[0]
    if 'nord' in kwargs:
        nord = kwargs['nord']
    else:
        nord = 3
    if 'bkptbin' in kwargs:
        bkptbin = kwargs['bkptbin']
    else:
        bkptbin = 1.2 * binsz
    if 'maxsep' in kwargs:
        maxsep = kwargs['maxsep']
    else:
        maxsep = 2.0 * binsz
    if inloglam.ndim == 1:
        #
        # Set specnum = 0 for all elements
        #
        npix = inloglam.shape[0]
        nspec = 1
        specnum = np.zeros(inloglam.shape,dtype=inloglam.dtype)
    else:
        nspec,npix = inloglam.shape
        specnum = np.tile(np.arange(nspec),npix).reshape(npix,nspec).transpose()
    #
    # Use fullcombmask for modifying the pixel masks in the original input files.
    #
    fullcombmask = np.zeros(npix)
    newflux = np.zeros(nfinalpix,dtype=inloglam.dtype)
    newmask = np.zeros(nfinalpix,dtype='i4')
    newivar = np.zeros(nfinalpix,dtype=inloglam.dtype)
    newdisp = np.zeros(nfinalpix,dtype=inloglam.dtype)
    newsky = np.zeros(nfinalpix,dtype=inloglam.dtype)
    newdispweight = np.zeros(nfinalpix,dtype=inloglam.dtype)
    if objivar is None:
        nonzero = np.arange(npix,dtype='i4')
        ngood = npix
    else:
        nonzero = (objivar > 0).nonzero()[0]
        ngood = len(nonzero)
    #
    # ormask is needed to create andmask
    #
    andmask = np.zeros(nfinalpix,dtype='i4')
    ormask = np.zeros(nfinalpix,dtype='i4')
    if ngood == 0:
        #
        # In this case of no good points, set the nodata bit everywhere.
        # Also if noplug is set in the first input bit-mask, assume it
        # should be set everywhere in the output bit masks.  No other bits
        # are set.
        #
        if 'verbose' in kwargs:
            print('No good points')
        bitval = pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','NODATA')
        if 'finalmask' in kwargs:
            bitval = bitval | (pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','NOPLUG') *
                (finalmask[0] & pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','NODATA')))
        andmask = andmask | bitval
        ormask = ormask | bitval
        return (newflux,newivar)
    else:
        #
        # Now let's break sorted wavelengths into groups where pixel
        # separations are larger than maxsep.
        #
        isort = nonzero[inloglam[nonzero].argsort()]
        wavesort = inloglam[isort]
        padwave = np.insert(wavesort,0,wavesort.min() - 2.0*maxsep)
        padwave = np.append(padwave,wavesort.max() + 2.0*maxsep)
        ig1 = ((padwave[1:ngood+1]-padwave[0:ngood]) > maxsep).nonzero()[0]
        ig2 = ((padwave[2:ngood+2]-padwave[1:ngood+1]) > maxsep).nonzero()[0]
        if ig1.size != ig2.size:
            raise ValueError('Grouping tricks did not work!')
        for igrp in range(ig1.size):
            ss = isort[ig1[igrp]:ig2[igrp]+1]
            if ss.size > 2:
                if objivar is None:
                    #
                    # Fit without variance
                    #
                    sset,bmask = pydl.pydlutils.bspline.iterfit(inloglam[ss],objflux[ss],
                        nord=nord,groupbadpix=True,requiren=1,bkspace=bkptbin,
                        silent=True)
                else:
                    #
                    # Fit with variance
                    #
                    sset,bmask = pydl.pydlutils.bspline.iterfit(inloglam[ss],objflux[ss],
                        invvar=objivar[ss],
                        nord=nord,groupbadpix=True,requiren=1,bkspace=bkptbin,
                        silent=True)
                if np.sum(np.absolute(sset.coeff)) == 0:
                    sset = None
                    bmask = np.zeros(len(ss))
                    if 'verbose' in kwargs:
                        print('WARNING: All B-spline coefficients have been set to zero!')
            else:
                bmask = np.zeros(len(ss))
                sset = None
                if 'verbose' in kwargs:
                    print('WARNING: Not enough data for B-spline fit!')
            inside = ((newloglam >= inloglam[ss].min()-EPS) &
                (newloglam <= inloglam[ss].max()+EPS)).nonzero()[0]
            #
            # It is possible for numinside to be zero, if the input data points
            # span an extremely small wavelength range, within which there are
            # no output wavelengths.
            #
            if sset is not None and len(inside) > 0:
                newflux[inside],bvalumask = sset.value(newloglam[inside])
                if bvalumask.any():
                    newmask[inside[bvalumask]] = 1
                if 'verbose' in kwargs:
                    print('Masked {0:d} of {1:d} pixels.'.format(bmask.sum()-bmask.size,bmask.size))
                #
                # Determine which pixels should be masked based upon the spline
                # fit. Set the combinerej bit.
                #
                ireplace = ~bmask
                if ireplace.any():
                    #
                    # The following would replace the original flux values of
                    # masked pixels with b-spline evaluations.
                    #
                    # objflux[ss[ireplace]] = sset.value(inloglam[ss[ireplace]])
                    #
                    # Set the inverse variance of these pixels to zero.
                    #
                    if objivar is not None:
                        objivar[ss[ireplace]] = 0.0
                        if 'verbose' in kwargs:
                            print('Replaced {0:d} pixels in objivar.'.format(len(ss[ireplace])))
                    if 'finalmask' in kwargs:
                        finalmask[ss[ireplace]] = (finalmask[ss[ireplace]] |
                            pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','COMBINEREJ'))
            fullcombmask[ss] = bmask
        #
        # Combine inverse variance and pixel masks.
        #
        # Start with all bits set in andmask
        #
        andmask[:] = -1
        for j in range(int(specnum.max())+1):
            these = specnum == j
            if these.any():
                inbetween = ((newloglam >= inloglam[these].min()) &
                    (newloglam <= inloglam[these].max()))
                if inbetween.any():
                    jnbetween = inbetween.nonzero()[0]
                    #
                    # Conserve inverse variance by doing a linear interpolation
                    # on that quantity.
                    #
                    result = np.interp(newloglam[jnbetween],
                        inloglam[these],objivar[these]*fullcombmask[these])
                    #
                    # Grow the fullcombmask below to reject any new sampling
                    # containing even a partial masked pixel.
                    #
                    smask = np.interp(newloglam[jnbetween],inloglam[these],
                        fullcombmask[these].astype(inloglam.dtype))
                    result *= smask >= (1.0 - EPS)
                    newivar[jnbetween] += result*newmask[jnbetween]
                lowside = np.floor((inloglam[these]-newloglam[0])/binsz).astype('i4')
                highside = lowside + 1
                if 'finalmask' in kwargs:
                    andmask[lowside] &= finalmask[these]
                    andmask[highside] &= finalmask[these]
                    ormask[lowside] |= finalmask[these]
                    ormask[highside] |= finalmask[these]
                #
                # Combine the dispersions + skies in the dumbest way possible
                # [sic].
                #
                if 'indisp' in kwargs:
                    newdispweight[jnbetween] += result
                    newdisp[jnbetween] += result * np.interp(newloglam[jnbetween],
                        inloglam[these],indisp[these])
                    newsky[jnbetween] += result * np.interp(newloglam[jnbetween],
                        inloglam[these],skyflux[these])
        if 'indisp' in kwargs:
            newdisp /= newdispweight + (newdispweight == 0)
            newsky /= newdispweight + (newdispweight == 0)
    #
    # Grow regions where 3 or more pixels are rejected together ???
    #
    # print newivar
    foo = smooth(newivar,3)
    # print foo
    # sys.exit(1)
    badregion = np.absolute(foo) < EPS
    if badregion.any():
        if 'verbose' in kwargs:
            print('WARNING: Growing bad pixel region, {0:d} pixels found.'.format(badregion.sum()))
        ibad = badregion.nonzero()[0]
        lowerregion = np.where(ibad-2 < 0,0,ibad-2)
        upperregion = np.where(ibad+2 > nfinalpix-1,nfinalpix-1,ibad+2)
        newivar[lowerregion] = 0.0
        newivar[upperregion] = 0.0
    #
    # Replace NaNs in combined spectra; this should really never happen.
    #
    inff = ((~np.isfinite(newflux)) | (~np.isfinite(newivar)))
    if inff.any():
        print('WARNING: {0:d} NaNs in combined spectra.'.format(inff.sum()))
        newflux[inff] = 0.0
        newivar[inff] = 0.0
    #
    # Interpolate over masked pixels, just for aesthetic purposoes.
    #
    goodpts = newivar > 0
    if 'aesthetics' in kwargs:
        amethod = kwargs['aesthetics']
    else:
        amethod = 'traditional'
    newflux = aesthetics(newflux,newivar,method=amethod)
    # if 'interpolate' in kwargs:
    #     newflux = pydlutils.image.djs_maskinterp(newflux,~goodpts,const=True)
    # else:
    #     newflux[~goodpts] = newflux[goodpts].mean()
    if goodpts.any():
        minglam = newloglam[goodpts].min()
        maxglam = newloglam[goodpts].max()
        ibad = ((newloglam < minglam) | (newloglam > maxglam))
        if ibad.any():
            ormask[ibad] |= pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','NODATA')
    #
    # Replace values of -1 in the andmask with 0.
    #
    andmask *= (andmask != -1)
    #
    # Copy the nodata bad pixels from the ormask to the andmask.
    #
    andmask |= ormask & pydl.pydlutils.sdss.sdss_flagval('SPPIXMASK','NODATA')
    return (newflux,newivar)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the spec1d directory of idlspec2d.
"""
from .findspec import findspec
from .hmf_gal import hmf_gal
from .hmf_solve import hmf_solve
from .latest_mjd import latest_mjd
from .number_of_fibers import number_of_fibers
from .pca_gal import pca_gal
from .pca_qso import pca_qso
from .pca_solve import pca_solve
from .pca_star import pca_star
from .plot_eig import plot_eig
from .readspec import readspec
from .skymask import skymask
from .spec_append import spec_append
from .spec_path import spec_path
from .wavevector import wavevector
#
# Used by findspec
#
findspec_cache = None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_find_contiguous():
    from numpy import array
    from .. import find_contiguous
    assert find_contiguous(array([0,1,1,1,0,1,1,0,1])) == [1,2,3]



#
#
#
def xyz_to_angles(x,y,z):
    """Convert Cartesion coordinates to spherical coordinates (r,phi,theta)

    This uses the standard mathematical definition of spherical coordinates.
    To convert theta to Dec, use Dec = 90 - theta.
    """
    import numpy as np
    r = np.sqrt(x*x + y*y + z*z)
    theta = np.rad2deg(np.arccos(z/r))
    phi = np.rad2deg(np.arctan2(y,x))
    w = phi < 0
    if w.sum() > 0:
        phi[w] += 360.0
    return (r,phi,theta)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
class chunks(object):
    """chunks class

    Functions for creating and manipulating spherical chunks are implemented
    as methods on this class.
    """
    import numpy as np
    from .. import PydlutilsException
    def __init__(self, ra, dec, minSize):
        """Init creates an object whose attributes are similar those created
        by the setchunks() function in the spheregroup library.
        """
        #
        # Save the value of minSize
        #
        self.minSize = minSize
        #
        # Find maximum and minimum dec (in degrees)
        #
        decMin = dec.min()
        decMax = dec.max()
        decRange = decMax - decMin
        #
        # Find the declination boundaries; make them an integer multiple of
        # minSize, with extra room (one cell) on the edges.
        #
        self.nDec = 3 + int(self.np.floor(decRange/minSize))
        decRange = minSize*float(self.nDec)
        decMin = decMin - 0.5*(decRange - decMax + decMin)
        decMax = decMin + decRange
        if decMin < -90.0 + 3.0*minSize:
            decMin = -90.0
        if decMax > 90.0 - 3.0*minSize:
            decMax = 90.0
        self.decBounds = decMin + (decMax-decMin)*self.np.arange(self.nDec+1,dtype='d')/float(self.nDec)
        #
        # Find ra offset which minimizes the range in ra (this should take care
        # of the case that ra crosses zero in some parts
        #
        if abs(self.decBounds[self.nDec]) > abs(self.decBounds[0]):
            cosDecMin = self.np.cos(self.np.deg2rad(self.decBounds[self.nDec]))
        else:
            cosDecMin = self.np.cos(self.np.deg2rad(self.decBounds[0]))
        if cosDecMin <= 0.0:
            raise self.PydlutilsException("cosDecMin={0:f} not positive in setchunks().".format(cosDecMin))
        self.raRange, self.raOffset = self.rarange(ra,minSize/cosDecMin)
        self.raMin, self.raMax = self.getraminmax(ra,self.raOffset)
        #
        # Isn't this redundant?
        #
        self.raRange = self.raMax - self.raMin
        #
        # For each declination slice, find the number of ra divisions
        # necessary and set them
        #
        self.raBounds = list()
        self.nRa = list()
        for i in range(self.nDec):
            #
            # Get maximum declination and its cosine
            #
            if abs(self.decBounds[i]) > abs(self.decBounds[i+1]):
                cosDecMin = self.np.cos(self.np.deg2rad(self.decBounds[i]))
            else:
                cosDecMin = self.np.cos(self.np.deg2rad(self.decBounds[i+1]))
            if cosDecMin <= 0.0:
                raise self.PydlutilsException("cosDecMin={0:f} not positive in setchunks().".format(cosDecMin))
            #
            # Get raBounds array for this declination array, leave an extra
            # cell on each end
            #
            self.nRa.append(3 + int(self.np.floor(cosDecMin*self.raRange/minSize)))
            raRangeTmp = minSize*float(self.nRa[i])/cosDecMin
            raMinTmp = self.raMin - 0.5*(raRangeTmp-self.raMax+self.raMin)
            raMaxTmp = raMinTmp + raRangeTmp
            #
            # If we cannot avoid the 0/360 point, embrace it
            #
            if ( raRangeTmp >= 360.0 or
                raMinTmp <= minSize/cosDecMin or
                raMaxTmp >= 360.0 - minSize/cosDecMin or
                abs(self.decBounds[i]) == 90.0 ):
                raMinTmp = 0.0
                raMaxTmp = 360.0
                raRangeTmp = 360.0
            if self.decBounds[i] == -90.0 or self.decBounds[i+1] == 90.0:
                self.nRa[i] = 1
            self.raBounds.append(raMinTmp +
                (raMaxTmp-raMinTmp)*self.np.arange(self.nRa[i]+1,dtype='d')/
                float(self.nRa[i]))
        #
        # Create an empty set of lists to hold the output of self.assign()
        #
        self.chunkList = [[list() for j in range(self.nRa[i])] for i in range(self.nDec)]
        #
        # nChunkMax will be the length of the largest list in chunkList
        # it is computed by chunks.assign()
        #
        self.nChunkMax = 0
        return
    def rarange(self,ra,minSize):
        """Finds the offset which yields the smallest raRange & returns both.
        """
        NRA = 6
        raRangeMin = 361.0
        raOffset = 0.0
        EPS = 1.0e-5
        for j in range(NRA):
            raMin,raMax = self.getraminmax(ra, 360.0*float(j)/float(NRA))
            raRange = raMax-raMin
            # print(minSize,raMin,raMax, raRange, raRangeMin)
            if (2.0*(raRange-raRangeMin)/(raRange+raRangeMin) < -EPS and
                raMin > minSize and raMax < 360.0 - minSize):
                # print(j)
                raRangeMin = raRange
                raOffset = 360.0*float(j)/float(NRA)
        # print(raRangeMin, raOffset)
        return (raRangeMin, raOffset)
    def getraminmax(self,ra,raOffset):
        """Utility function used by rarange.
        """
        currRa = self.np.fmod(ra + raOffset, 360.0)
        return (currRa.min(), currRa.max())
    def cosDecMin(self,i):
        """Frequently used utility function.
        """
        if abs(self.decBounds[i]) > abs(self.decBounds[i+1]):
            return self.np.cos(self.np.deg2rad(self.decBounds[i]))
        else:
            return self.np.cos(self.np.deg2rad(self.decBounds[i+1]))
    def assign(self, ra, dec, marginSize):
        """Take the objects and the chunks (already defined in the constructor)
        and assign the objects to the appropriate chunks, with some leeway
        given by the parameter marginSize.  Basically, at the end, each
        chunk should be associated with a list of the objects that belong
        to it.
        """
        if marginSize >= self.minSize:
            raise self.PydlutilsException("marginSize>=minSize ({0:f}={1:f}) in chunks.assign().".format(marginSize,self.minSize))
        chunkDone = [[False for j in range(self.nRa[i])] for i in range(self.nDec)]
        for i in range(ra.size):
            currRa = self.np.fmod(ra[i]+self.raOffset,360.0)
            try:
                raChunkMin, raChunkMax, decChunkMin, decChunkMax = self.getbounds(currRa,dec[i],marginSize)
            except self.PydlutilsException:
                continue
            #
            # Reset chunkDone.  This is silly, but is necessary to
            # reproduce the logic.
            #
            for decChunk in range(decChunkMin,decChunkMax+1):
                for raChunk in range(raChunkMin[decChunk-decChunkMin]-1,raChunkMax[decChunk-decChunkMin]+2):
                    if raChunk < 0:
                        currRaChunk = (raChunk+self.nRa[decChunk]) % self.nRa[decChunk]
                    elif raChunk > self.nRa[decChunk]-1:
                        currRaChunk = (raChunk-self.nRa[decChunk]) % self.nRa[decChunk]
                    else:
                        currRaChunk = raChunk
                    if currRaChunk >= 0 and currRaChunk <= self.nRa[decChunk]-1:
                        chunkDone[decChunk][currRaChunk] = False
            for decChunk in range(decChunkMin,decChunkMax+1):
                for raChunk in range(raChunkMin[decChunk-decChunkMin],raChunkMax[decChunk-decChunkMin]+1):
                    if raChunk < 0:
                        currRaChunk = (raChunk+self.nRa[decChunk]) % self.nRa[decChunk]
                    elif raChunk > self.nRa[decChunk]-1:
                        currRaChunk = (raChunk-self.nRa[decChunk]) % self.nRa[decChunk]
                    else:
                        currRaChunk = raChunk
                    if currRaChunk >= 0 and currRaChunk <= self.nRa[decChunk]-1:
                        if not chunkDone[decChunk][currRaChunk]:
                            self.chunkList[decChunk][currRaChunk].append(i)
                            #
                            # Update nChunkMax
                            #
                            if len(self.chunkList[decChunk][currRaChunk]) > self.nChunkMax:
                                self.nChunkMax = len(self.chunkList[decChunk][currRaChunk])
                            chunkDone[decChunk][currRaChunk] = True
        return
    def getbounds(self,ra,dec,marginSize):
        """Find the set of chunks a point (with margin) belongs to.
        """
        #
        # Find the declination slice without regard to marginSize
        #
        decChunkMin = int(self.np.floor((dec - self.decBounds[0])*
            float(self.nDec)/
            (self.decBounds[self.nDec]-self.decBounds[0])))
        decChunkMax = decChunkMin
        if decChunkMin < 0 or decChunkMin > self.nDec - 1:
            raise self.PydlutilsException("decChunkMin out of range in chunks.getbounds().")
        #
        # Set minimum and maximum bounds of dec
        #
        while dec-self.decBounds[decChunkMin] < marginSize and decChunkMin > 0:
            decChunkMin -= 1
        while self.decBounds[decChunkMax+1]-dec < marginSize and decChunkMax < self.nDec -1:
            decChunkMax += 1
        #
        # Find ra chunk bounds for each dec chunk
        #
        raChunkMin = self.np.zeros(decChunkMax-decChunkMin+1,dtype='i4')
        raChunkMax = self.np.zeros(decChunkMax-decChunkMin+1,dtype='i4')
        for i in range(decChunkMin,decChunkMax+1):
            cosDecMin = self.cosDecMin(i)
            raChunkMin[i-decChunkMin] = int(self.np.floor((ra - self.raBounds[i][0])*
                float(self.nRa[i])/
                (self.raBounds[i][self.nRa[i]] - self.raBounds[i][0])))
            raChunkMax[i-decChunkMin] = raChunkMin[i-decChunkMin]
            if raChunkMin[i-decChunkMin] < 0 or raChunkMin[i-decChunkMin] > self.nRa[i]-1:
                raise self.PydlutilsException("raChunkMin out of range in chunks.getbounds().")
            #
            # Set minimum and maximum bounds of ra
            #
            raCheck = raChunkMin[i-decChunkMin]
            keepGoing = True
            while keepGoing and raCheck > -1:
                if raCheck >= 0 and raCheck < self.nRa[i]:
                    keepGoing = (ra - self.raBounds[i][raCheck])*cosDecMin < marginSize
                else:
                    keepGoing = False
                if keepGoing:
                    raCheck -= 1
            raChunkMin[i-decChunkMin] = raCheck
            raCheck = raChunkMax[i-decChunkMin]
            keepGoing = True
            while keepGoing and raCheck < self.nRa[i]:
                if raCheck >= 0 and raCheck < self.nRa[i]:
                    keepGoing = (self.raBounds[i][raCheck+1]-ra)*cosDecMin < marginSize
                else:
                    keepGoing = False
                if keepGoing:
                    raCheck += 1
            raChunkMax[i-decChunkMin] = raCheck
        return (raChunkMin, raChunkMax, decChunkMin, decChunkMax)
    def get(self, ra, dec):
        """Find the chunk to which a given point belongs.
        """
        #
        # Find dec chunk
        #
        decChunk = int(self.np.floor((dec - self.decBounds[0])*
            float(self.nDec)/
            (self.decBounds[self.nDec]-self.decBounds[0])))
        #
        # Find ra chunk
        #
        if decChunk < self.nDec and decChunk >= 0:
            raChunk = int(self.np.floor((ra - self.raBounds[decChunk][0])*
                float(self.nRa[decChunk])/
                (self.raBounds[decChunk][self.nRa[decChunk]] - self.raBounds[decChunk][0])))
            if raChunk < 0 or raChunk > self.nRa[decChunk]-1:
                raise self.PydlutilsException("raChunk out of range in chunks.get()")
        else:
            raChunk = -1
        return (raChunk, decChunk)
    def friendsoffriends(self,ra,dec,linkSep):
        """Friends-of-friends using chunked data.
        """
        nPoints = ra.size
        inGroup = self.np.zeros(nPoints,dtype='i4') - 1
        #
        # mapGroups contains an equivalency mapping of groups.  mapGroup[i]=j
        # means i and j are actually the same group.  j<=i always, by design.
        # The largest number of groups you can get
        # (assuming linkSep < marginSize < minSize) is 9 times the number of
        # targets
        #
        mapGroups = self.np.zeros(9*nPoints,dtype='i4') - 1
        nMapGroups = 0
        for i in range(self.nDec):
            for j in range(self.nRa[i]):
                chunkGroup = self.chunkfriendsoffriends(ra,dec,self.chunkList[i][j],linkSep)
                for k in range(chunkGroup.nGroups):
                    minEarly = 9*nPoints
                    l = chunkGroup.firstGroup(k)
                    while l != -1:
                        if inGroup[self.chunkList[i][j][l]] != -1:
                            checkEarly = inGroup[self.chunkList[i][j][l]]
                            while mapGroups[checkEarly] != checkEarly:
                                checkEarly=mapGroups[checkEarly]
                            minEarly = min(minEarly,checkEarly)
                        else:
                            inGroup[self.chunkList[i][j][l]] = nMapGroups
                        l = chunkGroup.nextGroup(l)
                    if minEarly == 9*nPoints:
                        mapGroups[nMapGroups] = nMapGroups
                    else:
                        mapGroups[nMapGroups] = minEarly
                        l = chunkGroup.firstGroup(k)
                        while l != -1:
                            checkEarly = inGroup[self.chunkList[i][j][l]]
                            while mapGroups[checkEarly] != checkEarly:
                                tmpEarly=mapGroups[checkEarly]
                                mapGroups[checkEarly]=minEarly
                                checkEarly=tmpEarly
                            mapGroups[checkEarly]=minEarly
                            l=chunkGroup.nextGroup(l)
                    nMapGroups += 1
        #
        # Now all groups which are mapped to themselves are the real groups
        # Make sure the mappings are set up to go all the way down.
        #
        nGroups = 0
        for i in range(nMapGroups):
            if mapGroups[i] != -1:
                if mapGroups[i] == i:
                    mapGroups[i] = nGroups
                    nGroups += 1
                else:
                    mapGroups[i] = mapGroups[mapGroups[i]]
            else:
                raise self.PydlutilsException("MapGroups[{0:d}]={1:d} in chunks.friendsoffriends().".format(i,mapGroups[i]))
        for i in range(nPoints):
            inGroup[i] = mapGroups[inGroup[i]]
        firstGroup = self.np.zeros(nPoints,dtype='i4') - 1
        nextGroup = self.np.zeros(nPoints,dtype='i4') - 1
        multGroup = self.np.zeros(nPoints,dtype='i4')
        for i in range(nPoints-1,-1,-1):
            nextGroup[i] = firstGroup[inGroup[i]]
            firstGroup[inGroup[i]] = i
        for i in range(nGroups):
            j = firstGroup[i]
            while j != -1:
                multGroup[i] += 1
                j = nextGroup[j]
        return (inGroup, multGroup, firstGroup, nextGroup, nGroups)
    def chunkfriendsoffriends(self,ra,dec,chunkList,linkSep):
        """Does friends-of-friends on the ra, dec that are defined by
        chunkList.
        """
        from . import groups
        #
        # Convert ra, dec into something that can be digested by the
        # groups object.
        #
        x = self.np.deg2rad(self.np.vstack(ra[chunkList],dec[chunkList]))
        radLinkSep = self.np.deg2rad(linkSep)
        group = groups(x,radLinkSep,'sphereradec')
        return group



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
def decode_mixed(x):
    """Convert bytes in Numpy arrays into strings.  Leave other stuff alone.

    Parameters
    ----------
    x : object
        Input object.

    Returns
    -------
    decode_mixed : object
        If ``x`` has a ``decode()`` method, ``x.decode()`` will be returned.
        Otherwise ``x`` will be returned unchanged.
    """
    try:
        return x.decode()
    except:
        return x



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def read_fits_polygons(filename):
    """Read a "polygon" format FITS file.

    The main point of this is to extract the xcaps and cmcaps columns
    and replace them with caps.x and caps.cm, though this is really very silly
    & is only here so it's similar to IDL.
    """
    from astropy.io import fits
    hdulist = fits.open(filename)
    poly = hdulist[1].data
    hdulist.close()
    #poly['CAPS'] = {'CM':list(), 'X':list()}
    #for k in range(len(poly['NCAPS'])):
    #    poly['CAPS']['CM'].append(poly['CMCAPS'][k,0:poly['NCAPS'][k]])
    #    poly['CAPS']['X'].append(poly['XCAPS'][k].reshape((8,3))[0:poly['NCAPS'][k]])
    #del poly['CMCAPS']
    #del poly['XCAPS']
    return poly




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

import os
import numpy as np
from .. import yanny
from .. import write_ndarray_to_yanny
from . import YannyTestCase

class TestWriteNdarray(YannyTestCase):
    """Test class for ndarray to yanny conversions."""
    save_temp = False
    def test_write_single_ndarray_to_yanny(self):
        """Test the write_ndarray_to_yanny function."""
        mystruct = np.zeros((4,),
            dtype=[('ra','f8'), ('dec','f8'), ('mag','f4',(5,)),
                ('flags','i4'), ('new_flag','|S5')])
        mystruct['ra'] = np.array([10.0, 20.5, 30.75, 40.55],dtype=np.float64)
        mystruct['dec'] = np.array([-5.1234, -10.74832, 67.994523, 11.437281],dtype=np.float64)
        mystruct['mag'] = np.array([[0.0,1.0,2.0,3.0,4.0],
            [5.1,6.2,7.3,8.4,9.5],
            [22.123,23.95,22.6657,21.0286,22.9876],
            [13.54126,15.37456,14.52647,12.648640,12.0218]],dtype=np.float32)
        mystruct['flags'] = np.array([2**2, 2**4, 2**6, 2**8 + 2**3],dtype=np.int32)
        mystruct['new_flag'] = np.array(['FALSE','TRUE','TRUE','FALSE'],dtype='|S5')
        enums = {'new_flag':('BOOLEAN',('FALSE','TRUE'))}
        par = write_ndarray_to_yanny(self.temp('tempfile1.par'),mystruct,structnames='magnitudes',enums=enums)
        assert par['symbols']['enum'][0] == 'typedef enum {\n    FALSE,\n    TRUE\n} BOOLEAN;'
        assert par['symbols']['struct'][0] == 'typedef struct {\n    double ra;\n    double dec;\n    float mag[5];\n    int flags;\n    BOOLEAN new_flag;\n} MAGNITUDES;'
        for k,f in enumerate(('FALSE','TRUE','TRUE','FALSE')):
            assert par['MAGNITUDES']['new_flag'][k].decode() == f

    def test_write_multiple_ndarray_to_yanny(self):
        """Test the write_ndarray_to_yanny function."""
        mystruct = np.zeros((4,),
            dtype=[('ra','f8'), ('dec','f8'), ('mag','f4',(5,)),
                ('flags','i4'), ('new_flag','|S5')])
        mystruct['ra'] = np.array([10.0, 20.5, 30.75, 40.55],dtype=np.float64)
        mystruct['dec'] = np.array([-5.1234, -10.74832, 67.994523, 11.437281],dtype=np.float64)
        mystruct['mag'] = np.array([[0.0,1.0,2.0,3.0,4.0],
            [5.1,6.2,7.3,8.4,9.5],
            [22.123,23.95,22.6657,21.0286,22.9876],
            [13.54126,15.37456,14.52647,12.648640,12.0218]],dtype=np.float32)
        mystruct['flags'] = np.array([2**2, 2**4, 2**6, 2**8 + 2**3],dtype=np.int32)
        mystruct['new_flag'] = np.array(['FALSE','TRUE','TRUE','FALSE'],dtype='|S5')
        status = np.zeros((4,),dtype=[('timestamp','i8'),('state','S10')])
        status['timestamp'] = np.array([1382384327,1382384527,1382384727,1382384927],dtype='i8')
        status['state'] = np.array(['SUCCESS','SUCCESS','FAILURE','INCOMPLETE'],dtype='S10')
        enums = {'new_flag':('BOOLEAN',('FALSE','TRUE')),'state':('STATUS',('FAILURE','INCOMPLETE','SUCCESS'))}
        par = write_ndarray_to_yanny(self.temp('tempfile2.par'),(mystruct,status),structnames=('magnitudes','my_status'),enums=enums,comments=['This is a test','This is another test'])
        assert 'typedef enum {\n    FALSE,\n    TRUE\n} BOOLEAN;' in par['symbols']['enum']
        assert 'typedef enum {\n    FAILURE,\n    INCOMPLETE,\n    SUCCESS\n} STATUS;' in par['symbols']['enum']
        assert 'typedef struct {\n    double ra;\n    double dec;\n    float mag[5];\n    int flags;\n    BOOLEAN new_flag;\n} MAGNITUDES;' in par['symbols']['struct']
        assert 'typedef struct {\n    long timestamp;\n    STATUS state;\n} MY_STATUS;' in par['symbols']['struct']
        for k,f in enumerate(('FALSE','TRUE','TRUE','FALSE')):
            assert par['MAGNITUDES']['new_flag'][k].decode() == f



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_objid(run,camcol,field,objnum,rerun=301,skyversion=None):
    """Convert SDSS photometric identifiers into CAS-style ObjID.

    Bits are assigned in objid thus:

    ===== ========== ===============================================
    Bits  Name       Comment
    ===== ========== ===============================================
    63    empty      unassigned
    59-62 skyVersion resolved sky version (0-15)
    48-58 rerun      number of pipeline rerun
    32-47 run        run number
    29-31 camcol     camera column (1-6)
    28    firstField [is this the first field in segment?] 0 for now
    16-27 field      field number within run
    0-15  object     object number within field
    ===== ========== ===============================================

    Parameters
    ----------
    run, camcol, field, objnum : int or array of int
        Run, camcol, field and object number within field.  If arrays are passed,
        all must have the same length.
    rerun, skyversion : int or array of int, optional
        Rerun and skyversion usually don't change very much.  If supplied, make
        sure the size matches all the other values.

    Returns
    -------
    sdss_objid : ndarray of int64
        The ObjIDs of the objects.

    Raises
    ------
    ValueError
        If the sizes of the arrays don't match or if the array values are out of bounds.

    Notes
    -----
    firstField flag never set.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_objid
    >>> sdss_objid(3704,3,91,146)
    array([1237661382772195474])
    """
    from . import default_skyversion
    from numpy import array, int64, zeros
    if skyversion is None:
        skyversion = default_skyversion()
    if isinstance(run,int):
        #
        # Assume all inputs are integers & promote to arrays.
        #
        run = array([run],dtype=int64)
        camcol = array([camcol],dtype=int64)
        field = array([field],dtype=int64)
        objnum = array([objnum],dtype=int64)
        rerun = array([rerun],dtype=int64)
        skyversion = array([skyversion],dtype=int64)
        firstfield = array([0],dtype=int64)
    else:
        #
        # Check that all inputs have the same shape.
        #
        if rerun == 301:
            rerun = zeros(run.shape,dtype=int64) + 301
        if skyversion == default_skyversion():
            skyversion = zeros(run.shape,dtype=int64) + default_skyversion()
        firstfield = zeros(run.shape,dtype=int64)
        if run.shape != camcol.shape:
            raise ValueError("camcol.shape does not match run.shape!")
        if run.shape != field.shape:
            raise ValueError("field.shape does not match run.shape!")
        if run.shape != objnum.shape:
            raise ValueError("id.shape does not match run.shape!")
        if run.shape != rerun.shape:
            raise ValueError("id.shape does not match run.shape!")
        if run.shape != skyversion.shape:
            raise ValueError("id.shape does not match run.shape!")
    #
    # Check ranges of parameters
    #
    if ((skyversion < 0) | (skyversion >= 16)).any():
        raise ValueError("skyversion values are out-of-bounds!")
    if ((rerun < 0) | (rerun >= 2**11)).any():
        raise ValueError("rerun values are out-of-bounds!")
    if ((run < 0) | (run >= 2**16)).any():
        raise ValueError("run values are out-of-bounds!")
    if ((camcol < 1) | (camcol > 6)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((field < 0) | (field >= 2**12)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((objnum < 0) | (objnum >= 2**16)).any():
        raise ValueError("id values are out-of-bounds!")
    #
    # Compute the objid
    #
    objid = ((skyversion << 59) |
        (rerun << 48) |
        (run << 32) |
        (camcol << 29) |
        (firstfield << 28) |
        (field << 16) |
        (objnum))
    return objid
